import asyncio
from abc import ABC, abstractmethod
from asyncio import Future
from typing import Any
from uuid import UUID

from uagents_core.models import Model

from uagents.types import JsonStr, MsgInfo, RestMethod

PendingResponseKey = tuple[str, str, UUID]


class Sink(ABC):
    """
    Abstract base class for sinks that handle messages.
    """

    @abstractmethod
    async def handle_message(
        self, sender: str, schema_digest: str, message: JsonStr, session: UUID
    ):
        raise NotImplementedError

    @abstractmethod
    async def handle_rest(
        self, method: RestMethod, endpoint: str, message: Model | None
    ):
        raise NotImplementedError


class Dispatcher:
    """
    Dispatches incoming messages to internal sinks.
    """

    def __init__(self):
        self._sinks: dict[str, set[Sink]] = {}
        self._pending_responses: dict[PendingResponseKey, Future[MsgInfo]] = {}

    @property
    def sinks(self) -> dict[str, set[Sink]]:
        return self._sinks

    @property
    def pending_responses(self) -> dict[PendingResponseKey, Future[MsgInfo]]:
        return self._pending_responses

    def register_pending_response(self, sender: str, destination: str, session: UUID):
        self._pending_responses[(sender, destination, session)] = (
            asyncio.get_event_loop().create_future()
        )

    def cancel_pending_response(self, sender: str, destination: str, session: UUID):
        key: tuple[str, str, UUID] = (sender, destination, session)
        if key in self._pending_responses:
            del self._pending_responses[key]

    async def wait_for_response(
        self, sender: str, destination: str, session: UUID, timeout: float
    ) -> MsgInfo | None:
        key: tuple[str, str, UUID] = (sender, destination, session)
        try:
            response = await asyncio.wait_for(self._pending_responses[key], timeout)
        except asyncio.TimeoutError:
            response = None
        del self._pending_responses[key]
        return response

    def dispatch_pending_response(
        self,
        sender: str,
        destination: str,
        session: UUID,
        schema_digest: str,
        message: JsonStr,
    ) -> bool:
        key = (destination, sender, session)
        response = MsgInfo(message=message, sender=sender, schema_digest=schema_digest)
        if key in self._pending_responses:
            self._pending_responses[key].set_result(response)
            return True
        return False

    def register(self, address: str, sink: Sink):
        destinations = self._sinks.get(address, set())
        destinations.add(sink)
        self._sinks[address] = destinations

    def unregister(self, address: str, sink: Sink):
        destinations = self._sinks.get(address, set())
        destinations.discard(sink)
        if len(destinations) == 0:
            del self._sinks[address]
            return
        self._sinks[address] = destinations

    def contains(self, address: str) -> bool:
        return address in self._sinks

    async def dispatch_msg(
        self,
        sender: str,
        destination: str,
        schema_digest: str,
        message: JsonStr,
        session: UUID,
    ) -> None:
        if self.dispatch_pending_response(
            sender, destination, session, schema_digest, message
        ):
            return
        for handler in self._sinks.get(destination, set()):
            await handler.handle_message(sender, schema_digest, message, session)

    async def dispatch_rest(
        self,
        destination: str,
        method: RestMethod,
        endpoint: str,
        message: Model | None,
    ) -> dict[str, Any] | Model | None:
        for handler in self._sinks.get(destination, set()):
            return await handler.handle_rest(method, endpoint, message)


dispatcher = Dispatcher()



import asyncio
import contextlib
import json
import logging
import time
from datetime import datetime
from typing import Any

import aiohttp
import grpc
from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.tx import TxFee
from cosmpy.aerial.wallet import LocalWallet
from cosmpy.crypto.address import Address
from pydantic import BaseModel
from uagents_core.identity import Identity, parse_identifier
from uagents_core.registration import (
    AgentRegistrationAttestation,
    AgentRegistrationPolicy,
    BatchRegistrationPolicy,
    VerifiableModel,
)
from uagents_core.types import AgentEndpoint, AgentInfo

from uagents.config import (
    ALMANAC_API_MAX_RETRIES,
    ALMANAC_API_TIMEOUT_SECONDS,
    ALMANAC_API_URL,
    ALMANAC_REGISTRATION_WAIT,
    REGISTRATION_UPDATE_INTERVAL_SECONDS,
)
from uagents.crypto import sign_registration
from uagents.network import (
    DEFAULT_REGISTRATION_TIMEOUT_BLOCKS,
    AlmanacContract,
    AlmanacContractRecord,
    InsufficientFundsError,
    RetryDelayFunc,
    add_testnet_funds,
    default_exp_backoff,
)


class AgentRegistrationAttestationBatch(BaseModel):
    attestations: list[AgentRegistrationAttestation]


class AgentStatusUpdate(VerifiableModel):
    is_active: bool


def coerce_metadata_to_str(
    metadata: dict[str, Any] | None,
) -> dict[str, str | list[str] | dict[str, str]] | None:
    """Step through the metadata and convert any non-string values to strings."""
    if metadata is None:
        return None
    out = {}
    for key, val in metadata.items():
        if isinstance(val, dict):
            out[key] = {
                k: v if isinstance(v, str) else json.dumps(v) for k, v in val.items()
            }
        else:
            out[key] = val if isinstance(val, str) else json.dumps(val)
    return out


def extract_geo_metadata(metadata: dict[str, Any] | None) -> dict[str, Any] | None:
    """Extract geo-location metadata from the metadata dictionary."""
    if metadata is None:
        return None
    return {k: v for k, v in metadata.items() if k == "geolocation"}


async def almanac_api_post(
    url: str,
    data: BaseModel,
    *,
    timeout: float | None = None,
    max_retries: int | None = None,
    retry_delay: RetryDelayFunc | None = None,
) -> bool:
    """Send a POST request to the Almanac API."""
    timeout_seconds = timeout or ALMANAC_API_TIMEOUT_SECONDS
    num_retries = max_retries or ALMANAC_API_MAX_RETRIES
    retry_delay_func = retry_delay or default_exp_backoff

    async with aiohttp.ClientSession() as session:
        for retry in range(num_retries):
            try:
                async with session.post(
                    url=url,
                    headers={"content-type": "application/json"},
                    data=data.model_dump_json(),
                    timeout=aiohttp.ClientTimeout(total=timeout_seconds),
                ) as resp:
                    resp.raise_for_status()
                    return True
            except (aiohttp.ClientError, asyncio.exceptions.TimeoutError) as e:
                if retry + 1 >= num_retries:
                    raise e

                await asyncio.sleep(retry_delay_func(retry))
    return False


class AlmanacApiRegistrationPolicy(AgentRegistrationPolicy):
    def __init__(
        self,
        *,
        almanac_api: str | None = None,
        timeout: float | None = None,
        max_retries: int = ALMANAC_API_MAX_RETRIES,
        retry_delay: RetryDelayFunc | None = None,
        logger: logging.Logger | None = None,
    ):
        self._almanac_api = almanac_api or ALMANAC_API_URL
        self._timeout = timeout or ALMANAC_API_TIMEOUT_SECONDS
        self._max_retries = max_retries
        self._logger = logger or logging.getLogger(__name__)
        self._retry_delay = retry_delay or default_exp_backoff
        self._last_successful_registration: datetime | None = None

    @property
    def last_successful_registration(self) -> datetime | None:
        return self._last_successful_registration

    async def register(
        self,
        agent_identifier: str,
        identity: Identity,
        protocols: list[str],
        endpoints: list[AgentEndpoint],
        metadata: dict[str, Any] | None = None,
    ):
        # create the attestation
        attestation = AgentRegistrationAttestation(
            agent_identifier=agent_identifier,
            protocols=protocols,
            endpoints=endpoints,
            metadata=coerce_metadata_to_str(extract_geo_metadata(metadata)),
        )

        # sign the attestation
        attestation.sign(identity)

        try:
            success = await almanac_api_post(
                url=f"{self._almanac_api}/agents",
                data=attestation,
                timeout=self._timeout,
                max_retries=self._max_retries,
                retry_delay=self._retry_delay,
            )
            if success:
                self._logger.info("Registration on Almanac API successful")
                self._last_successful_registration = datetime.now()
            else:
                self._logger.warning("Registration on Almanac API failed")
        except Exception:
            self._logger.warning("Registration on Almanac API failed")


class BatchAlmanacApiRegistrationPolicy(BatchRegistrationPolicy):
    def __init__(
        self,
        almanac_api: str | None = None,
        logger: logging.Logger | None = None,
        timeout: float | None = None,
        max_retries: int = ALMANAC_API_MAX_RETRIES,
        retry_delay: RetryDelayFunc | None = None,
    ):
        self._almanac_api = almanac_api or ALMANAC_API_URL
        self._attestations: list[AgentRegistrationAttestation] = []
        self._logger = logger or logging.getLogger(__name__)
        self._timeout = timeout or ALMANAC_API_TIMEOUT_SECONDS
        self._max_retries = max_retries
        self._retry_delay = retry_delay or default_exp_backoff
        self._last_successful_registration: datetime | None = None

    @property
    def last_successful_registration(self) -> datetime | None:
        return self._last_successful_registration

    def add_agent(self, agent_info: AgentInfo, identity: Identity):
        attestation = AgentRegistrationAttestation(
            agent_identifier=f"{agent_info.prefix}://{agent_info.address}",
            protocols=list(agent_info.protocols),
            endpoints=agent_info.endpoints,
            metadata=coerce_metadata_to_str(extract_geo_metadata(agent_info.metadata)),
        )
        attestation.sign(identity)
        self._attestations.append(attestation)

    async def register(self):
        if not self._attestations:
            return
        attestations = AgentRegistrationAttestationBatch(
            attestations=self._attestations
        )

        try:
            success = await almanac_api_post(
                url=f"{self._almanac_api}/agents/batch",
                data=attestations,
                timeout=self._timeout,
                max_retries=self._max_retries,
                retry_delay=self._retry_delay,
            )
            if success:
                self._logger.info("Batch registration on Almanac API successful")
                self._last_successful_registration = datetime.now()
            else:
                self._logger.warning("Batch registration on Almanac API failed")
        except Exception:
            self._logger.warning("Batch registration on Almanac API failed")


class LedgerBasedRegistrationPolicy(AgentRegistrationPolicy):
    def __init__(
        self,
        ledger: LedgerClient,
        wallet: LocalWallet,
        almanac_contract: AlmanacContract,
        testnet: bool,
        *,
        tx_fee: TxFee | None = None,
        timeout_blocks: int = DEFAULT_REGISTRATION_TIMEOUT_BLOCKS,
        logger: logging.Logger | None = None,
    ):
        self._wallet = wallet
        self._ledger = ledger
        self._testnet = testnet
        self._almanac_contract = almanac_contract
        self._registration_fee = almanac_contract.get_registration_fee(wallet.address())
        self._logger = logger or logging.getLogger(__name__)
        self._broadcast_retries: int | None = None
        self._broadcast_retry_delay: RetryDelayFunc | None = None
        self._poll_retries: int | None = None
        self._poll_retry_delay: RetryDelayFunc | None = None
        self._last_successful_registration: datetime | None = None
        self._timeout_blocks = timeout_blocks
        self._tx_fee = tx_fee

    @property
    def last_successful_registration(self) -> datetime | None:
        return self._last_successful_registration

    @property
    def broadcast_retries(self) -> int | None:
        return self._broadcast_retries

    @broadcast_retries.setter
    def broadcast_retries(self, value: int | None) -> None:
        self._broadcast_retries = value

    @property
    def broadcast_retry_delay(self) -> RetryDelayFunc | None:
        return self._broadcast_retry_delay

    @broadcast_retry_delay.setter
    def broadcast_retry_delay(self, value: RetryDelayFunc | None) -> None:
        self._broadcast_retry_delay = value

    @property
    def poll_retries(self) -> int | None:
        return self._poll_retries

    @poll_retries.setter
    def poll_retries(self, value: int | None) -> None:
        self._poll_retries = value

    @property
    def poll_retry_delay(self) -> RetryDelayFunc | None:
        return self._poll_retry_delay

    @poll_retry_delay.setter
    def poll_retry_delay(self, value: RetryDelayFunc | None) -> None:
        self._poll_retry_delay = value

    async def register(
        self,
        agent_identifier: str,
        identity: Identity,
        protocols: list[str],
        endpoints: list[AgentEndpoint],
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """
        Register the agent on the Almanac contract if registration is about to expire or
        the registration data has changed.
        """
        _, _, agent_address = parse_identifier(agent_identifier)

        if (
            not self._almanac_contract.is_registered(agent_address)
            or self._almanac_contract.get_expiry(agent_address)
            < REGISTRATION_UPDATE_INTERVAL_SECONDS
            or endpoints != self._almanac_contract.get_endpoints(agent_address)
            or protocols != self._almanac_contract.get_protocols(agent_address)
        ):
            if self._get_balance() < self._registration_fee:
                self._logger.warning(
                    "I do not have enough funds to register on Almanac contract"
                )
                if self._testnet:
                    add_testnet_funds(str(self._wallet.address()))
                    self._logger.info(
                        f"Adding testnet funds to {self._wallet.address()}"
                    )
                else:
                    self._logger.info(
                        f"Send funds to wallet address: {self._wallet.address()}"
                    )
                raise InsufficientFundsError()

            self._logger.info("Registering on almanac contract...")

            current_time = int(time.time()) - ALMANAC_REGISTRATION_WAIT

            signature = self._sign_registration(identity, current_time)
            try:
                await self._almanac_contract.register(
                    ledger=self._ledger,
                    wallet=self._wallet,
                    agent_address=agent_address,
                    protocols=protocols,
                    endpoints=endpoints,
                    signature=signature,
                    current_time=current_time,
                    broadcast_retries=self._broadcast_retries,
                    broadcast_retry_delay=self._broadcast_retry_delay,
                    poll_retries=self._poll_retries,
                    poll_retry_delay=self._poll_retry_delay,
                    tx_fee=self._tx_fee,
                    timeout_blocks=self._timeout_blocks,
                )
                self._logger.info("Registering on almanac contract...complete")
                self._last_successful_registration = datetime.now()

            except RuntimeError as e:
                self._logger.warning(
                    "Registering on almanac contract...failed (will retry later)"
                )
                self._logger.debug(e)
            except grpc.RpcError as e:
                self._logger.warning(
                    "Registering on almanac contract...failed (will retry later)"
                )
                self._logger.debug(e)

        else:
            self._logger.info("Almanac contract registration is up to date!")

    def _get_balance(self) -> int:
        return self._ledger.query_bank_balance(Address(self._wallet.address()))

    def _sign_registration(self, identity: Identity, timestamp: int) -> str:
        """
        Sign the registration data for Almanac contract.

        Args:
            timestamp (int): The timestamp for the registration.

        Returns:
            str: The signature of the registration data.

        Raises:
            AssertionError: If the Almanac contract address is None.
        """
        assert self._almanac_contract.address is not None

        return sign_registration(
            identity=identity,
            contract_address=str(self._almanac_contract.address),
            timestamp=timestamp,
            wallet_address=str(self._wallet.address()),
        )


class BatchLedgerRegistrationPolicy(BatchRegistrationPolicy):
    def __init__(
        self,
        ledger: LedgerClient,
        wallet: LocalWallet,
        almanac_contract: AlmanacContract,
        testnet: bool,
        *,
        logger: logging.Logger | None = None,
        tx_fee: TxFee | None = None,
        timeout_blocks: int = DEFAULT_REGISTRATION_TIMEOUT_BLOCKS,
    ):
        self._ledger = ledger
        self._wallet = wallet
        self._almanac_contract = almanac_contract
        self._testnet = testnet
        self._registration_fee = almanac_contract.get_registration_fee(wallet.address())
        self._logger = logger or logging.getLogger(__name__)
        self._records: list[AlmanacContractRecord] = []
        self._identities: dict[str, Identity] = {}

        self._broadcast_retries: int | None = None
        self._broadcast_retry_delay: RetryDelayFunc | None = None
        self._poll_retries: int | None = None
        self._poll_retry_delay: RetryDelayFunc | None = None
        self._last_successful_registration: datetime | None = None
        self._tx_fee: TxFee | None = None
        self._timeout_blocks = timeout_blocks

    @property
    def last_successful_registration(self) -> datetime | None:
        return self._last_successful_registration

    @property
    def broadcast_retries(self) -> int | None:
        return self._broadcast_retries

    @broadcast_retries.setter
    def broadcast_retries(self, value: int | None) -> None:
        self._broadcast_retries = value

    @property
    def broadcast_retry_delay(self) -> RetryDelayFunc | None:
        return self._broadcast_retry_delay

    @broadcast_retry_delay.setter
    def broadcast_retry_delay(self, value: RetryDelayFunc | None) -> None:
        self._broadcast_retry_delay = value

    @property
    def poll_retries(self) -> int | None:
        return self._poll_retries

    @poll_retries.setter
    def poll_retries(self, value: int | None) -> None:
        self._poll_retries = value

    @property
    def poll_retry_delay(self) -> RetryDelayFunc | None:
        return self._poll_retry_delay

    @poll_retry_delay.setter
    def poll_retry_delay(self, value: RetryDelayFunc | None) -> None:
        self._poll_retry_delay = value

    def add_agent(self, agent_info: AgentInfo, identity: Identity) -> None:
        agent_record = AlmanacContractRecord(
            address=agent_info.address,
            prefix=agent_info.prefix,
            protocols=agent_info.protocols,
            endpoints=agent_info.endpoints,
            contract_address=str(self._almanac_contract.address),
            sender_address=str(self._wallet.address()),
            agent_type=agent_info.agent_type,
        )
        self._records.append(agent_record)
        self._identities[agent_info.address] = identity

    def _get_balance(self) -> int:
        return self._ledger.query_bank_balance(Address(self._wallet.address()))

    async def register(self) -> None:
        self._logger.info("Registering agents on Almanac contract...")
        for record in self._records:
            record.sign(self._identities[record.address])

        if self._get_balance() < self._registration_fee * len(self._records):
            self._logger.warning(
                f"I do not have enough funds to register {len(self._records)} "
                "agents on Almanac contract"
            )
            if self._testnet:
                add_testnet_funds(str(self._wallet.address()))
                self._logger.info(f"Adding testnet funds to {self._wallet.address()}")
            else:
                self._logger.info(
                    f"Send funds to wallet address: {self._wallet.address()}"
                )
            raise InsufficientFundsError()

        try:
            await self._almanac_contract.register_batch(
                ledger=self._ledger,
                wallet=self._wallet,
                agent_records=self._records,
                broadcast_retries=self._broadcast_retries,
                broadcast_retry_delay=self._broadcast_retry_delay,
                poll_retries=self._poll_retries,
                poll_retry_delay=self._poll_retry_delay,
                tx_fee=self._tx_fee,
                timeout_blocks=self._timeout_blocks,
            )

            self._logger.info("Registering agents on Almanac contract...complete")
            self._last_successful_registration = datetime.now()

        except RuntimeError as e:
            self._logger.warning(
                "Registering on almanac contract...failed (will retry later)"
            )
            self._logger.debug(e)
        except grpc.RpcError as e:
            self._logger.warning(
                "Registering on almanac contract...failed (will retry later)"
            )
            self._logger.debug(e)


class DefaultRegistrationPolicy(AgentRegistrationPolicy):
    def __init__(
        self,
        ledger: LedgerClient,
        wallet: LocalWallet,
        almanac_contract: AlmanacContract | None,
        testnet: bool,
        *,
        logger: logging.Logger | None = None,
        almanac_api: str | None = None,
    ):
        self._logger = logger or logging.getLogger(__name__)
        self._api_policy = AlmanacApiRegistrationPolicy(
            almanac_api=almanac_api, logger=logger
        )
        if almanac_contract is None:
            self._ledger_policy = None
        else:
            self._ledger_policy = LedgerBasedRegistrationPolicy(
                ledger, wallet, almanac_contract, testnet, logger=logger
            )

    async def register(
        self,
        agent_identifier: str,
        identity: Identity,
        protocols: list[str],
        endpoints: list[AgentEndpoint],
        metadata: dict[str, Any] | None = None,
    ) -> None:
        # prefer the API registration policy as it is faster
        try:
            await self._api_policy.register(
                agent_identifier, identity, protocols, endpoints, metadata
            )
        except Exception as e:
            self._logger.warning(
                f"Failed to register on Almanac API: {e.__class__.__name__}"
            )
            self._logger.debug(e)

        if self._ledger_policy is None:
            self._logger.info(
                "No Ledger available. Skipping registration on Almanac contract."
            )
            return

        # schedule the ledger registration
        try:
            await self._ledger_policy.register(
                agent_identifier, identity, protocols, endpoints, metadata
            )
        except InsufficientFundsError:
            self._logger.warning(
                "Failed to register on Almanac contract due to insufficient funds"
            )
            raise
        except Exception as e:
            self._logger.error(
                f"Failed to register on Almanac contract: {e.__class__.__name__}"
            )
            self._logger.debug(e)
            raise


async def update_agent_status(status: AgentStatusUpdate, almanac_api: str):
    _, _, agent_address = parse_identifier(status.agent_identifier)

    with contextlib.suppress(Exception):
        await almanac_api_post(
            url=f"{almanac_api}/agents/{agent_address}/status",
            data=status,
        )


class DefaultBatchRegistrationPolicy(BatchRegistrationPolicy):
    def __init__(
        self,
        ledger: LedgerClient,
        wallet: LocalWallet | None = None,
        almanac_contract: AlmanacContract | None = None,
        testnet: bool = True,
        *,
        logger: logging.Logger | None = None,
        almanac_api: str | None = None,
    ):
        self._logger = logger or logging.getLogger(__name__)
        self._api_policy = BatchAlmanacApiRegistrationPolicy(
            almanac_api=almanac_api, logger=logger
        )

        if almanac_contract is None or wallet is None:
            self._ledger_policy = None
        else:
            self._ledger_policy = BatchLedgerRegistrationPolicy(
                ledger, wallet, almanac_contract, testnet, logger=logger
            )

    def add_agent(self, agent_info: AgentInfo, identity: Identity) -> None:
        self._api_policy.add_agent(agent_info, identity)
        if self._ledger_policy is not None:
            self._ledger_policy.add_agent(agent_info, identity)

    async def register(self) -> None:
        # prefer the API registration policy as it is faster
        try:
            await self._api_policy.register()
        except Exception as e:
            self._logger.warning(
                f"Failed to batch register on Almanac API: {e.__class__.__name__}"
            )
            self._logger.debug(e)

        if self._ledger_policy is None:
            return

        # schedule the ledger registration
        try:
            await self._ledger_policy.register()
        except InsufficientFundsError:
            self._logger.warning(
                "Failed to batch register on Almanac contract due to insufficient funds"
            )
            raise
        except Exception as e:
            self._logger.error(
                f"Failed to batch register on Almanac contract: {e.__class__.__name__}"
            )
            self._logger.debug(e)
            raise



"""Agent Context and Message Handling"""

import asyncio
import logging
import uuid
from abc import ABC, abstractmethod
from collections.abc import Callable
from time import time
from typing import TYPE_CHECKING, Any

import requests
from cosmpy.aerial.client import LedgerClient
from pydantic.v1 import ValidationError
from uagents_core.envelope import Envelope
from uagents_core.identity import parse_identifier
from uagents_core.models import ERROR_MESSAGE_DIGEST, ErrorMessage, Model
from uagents_core.types import DeliveryStatus, MsgStatus

from uagents.communication import dispatch_local_message
from uagents.config import (
    ALMANAC_API_URL,
    DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    DEFAULT_SEARCH_LIMIT,
)
from uagents.dispatch import dispatcher
from uagents.resolver import Resolver
from uagents.storage import KeyValueStore
from uagents.types import EnvelopeHistory, EnvelopeHistoryEntry, JsonStr, MsgInfo
from uagents.utils import log

if TYPE_CHECKING:
    from uagents.agent import AgentRepresentation
    from uagents.communication import Dispenser
    from uagents.protocol import Protocol


class Context(ABC):
    """
    Represents the context in which messages are handled and processed.

    Properties:
        agent (AgentRepresentation): The agent representation associated with the context.
        storage (KeyValueStore): The key-value store for storage operations.
        ledger (LedgerClient): The client for interacting with the blockchain ledger.
        logger (logging.Logger): The logger instance.
        session (uuid.UUID): The session UUID associated with the context.

    Methods:
        get_agents_by_protocol(protocol_digest, limit, logger): Retrieve a list of agent addresses
            using a specific protocol digest.
        broadcast(destination_protocol, message, limit, timeout): Broadcast a message
            to agents with a specific protocol.
        session_history: Get the message history associated with the context session.
        send(destination, message, timeout): Send a message to a destination.
        send_raw(destination, json_message, schema_digest, message_type, timeout):
            Send a message with the provided schema digest to a destination.
    """

    @property
    @abstractmethod
    def agent(self) -> "AgentRepresentation":
        """
        Get the agent representation associated with the context.

        Returns:
            AgentRepresentation: The agent representation.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def storage(self) -> KeyValueStore:
        """
        Get the key-value store associated with the context.

        Returns:
            KeyValueStore: The key-value store.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def ledger(self) -> LedgerClient:
        """
        Get the ledger client associated with the context.

        Returns:
            LedgerClient: The ledger client.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def logger(self) -> logging.Logger:
        """
        Get the logger instance associated with the context.

        Returns:
            logging.Logger: The logger instance.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def session(self) -> uuid.UUID:
        """
        Get the session UUID associated with the context.

        Returns:
            uuid.UUID: The session UUID.
        """
        raise NotImplementedError

    @abstractmethod
    def get_agents_by_protocol(
        self,
        protocol_digest: str,
        limit: int = DEFAULT_SEARCH_LIMIT,
        logger: logging.Logger | None = None,
    ) -> list[str]:
        """Retrieve a list of agent addresses using a specific protocol digest.

        This method queries the Almanac API to retrieve a list of agent addresses
        that are associated with a given protocol digest. The list can be optionally
        limited to a specified number of addresses.

        Args:
            protocol_digest (str): The protocol digest to search for, starting with "proto:".
            limit (int, optional): The maximum number of agent addresses to return.

        Returns:
            list[str]: A list of agent addresses using the specified protocol digest.
        """
        raise NotImplementedError

    @abstractmethod
    async def broadcast(
        self,
        destination_protocol: str,
        message: Model,
        limit: int = DEFAULT_SEARCH_LIMIT,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    ) -> list[MsgStatus]:
        """Broadcast a message to agents with a specific protocol.

        This asynchronous method broadcasts a given message to agents associated
        with a specific protocol. The message is sent to multiple agents concurrently.
        The schema digest of the message is used for verification.

        Args:
            destination_protocol (str): The protocol to filter agents by.
            message (Model): The message to broadcast.
            limit (int, optional): The maximum number of agents to send the message to.
            timeout (int, optional): The timeout for sending each message.

        Returns:
            list[MsgStatus]: A list of message delivery statuses.
        """
        raise NotImplementedError

    @abstractmethod
    def session_history(self) -> list[EnvelopeHistoryEntry] | None:
        """
        Get the message history associated with the context session.

        Returns:
            list[EnvelopeHistoryEntry] | None: The message history.
        """
        pass

    @abstractmethod
    async def send(
        self,
        destination: str,
        message: Model,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    ) -> MsgStatus:
        """
        Send a message to the specified destination.

        Args:
            destination (str): The destination address to send the message to.
            message (Model): The message to be sent.
            timeout (int, optional): The timeout for sending the message, in seconds.

        Returns:
            MsgStatus: The delivery status of the message.
        """
        raise NotImplementedError

    @abstractmethod
    async def send_raw(
        self,
        destination: str,
        message_schema_digest: str,
        message_body: JsonStr,
        sync: bool = False,
        wait_for_response: bool = False,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
        protocol_digest: str | None = None,
        queries: dict[str, asyncio.Future] | None = None,
    ) -> MsgStatus:
        """
        Send a message to the specified destination where the message body and
        message schema digest are sent separately.

        Args:
            destination (str): The destination address to send the message to.
            message_schema_digest (str): The schema digest of the message to be sent.
            message_body (JsonStr): The JSON-encoded message body to be sent.
            sync (bool): Whether to send the message synchronously or asynchronously.
            wait_for_response (bool): Whether to wait for a response to the message.
            timeout (int, optional): The optional timeout for sending the message, in seconds.
            protocol_digest (str, optional): The protocol digest of the message to be sent.
            queries (dict[str, asyncio.Future] | None): The dictionary of queries to resolve.

        Returns:
            MsgStatus: The delivery status of the message.
        """
        raise NotImplementedError

    @abstractmethod
    async def send_and_receive(
        self,
        destination: str,
        message: Model,
        response_type: type[Model] | set[type[Model]],
        sync: bool = False,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    ) -> tuple[Model | None, MsgStatus]:
        """
        Send a message to the specified destination and receive a response.

        Args:
            destination (str): The destination address to send the message to.
            message (Model): The message to be sent.
            response_type (type[Model] | set[type[Model]]): The type(s) of the response message.
            sync (bool): Whether to send the message synchronously or asynchronously.
            timeout (int): The timeout for sending the message, in seconds.

        Returns:
            tuple[Model | None, MsgStatus]: The response message if received and delivery status
        """
        raise NotImplementedError

    @abstractmethod
    async def send_wallet_message(
        self,
        destination: str,
        text: str,
        msg_type: int = 1,
    ):
        """
        Send a message to the wallet of the specified destination.

        Args:
            destination (str): The destination address to send the message to.
            text (str): The text of the message to be sent.
            msg_type (int): The type of the message to be sent.

        Returns:
            None
        """
        raise NotImplementedError


class InternalContext(Context):
    """Represents the agent internal context for proactive behaviour."""

    def __init__(
        self,
        agent: "AgentRepresentation",
        storage: KeyValueStore,
        ledger: LedgerClient,
        resolver: Resolver,
        dispenser: "Dispenser",
        session: uuid.UUID | None = None,
        interval_messages: set[str] | None = None,
        wallet_messaging_client: Any | None = None,
        message_history: EnvelopeHistory | None = None,
        logger: logging.Logger | None = None,
    ):
        self._agent = agent
        self._storage = storage
        self._ledger = ledger
        self._resolver = resolver
        self._dispenser = dispenser
        self._logger = logger
        self._session = session or uuid.uuid4()
        self._interval_messages = interval_messages
        self._wallet_messaging_client = wallet_messaging_client
        self._message_history = message_history
        self._outbound_messages: dict[str, list[tuple[JsonStr, str]]] = {}

    @property
    def agent(self) -> "AgentRepresentation":
        return self._agent

    @property
    def storage(self) -> KeyValueStore:
        return self._storage

    @property
    def ledger(self) -> LedgerClient:
        return self._ledger

    @property
    def logger(self) -> logging.Logger | None:
        return self._logger

    @property
    def session(self) -> uuid.UUID:
        """
        Get the session UUID associated with the context.

        Returns:
            uuid.UUID: The session UUID.
        """
        return self._session

    @property
    def outbound_messages(self) -> dict[str, list[tuple[JsonStr, str]]]:
        """
        Get the dictionary of outbound messages associated with the context.

        Returns:
            dict[str, list[tuple[JsonStr, str]]]: The dictionary of outbound messages.
        """
        return self._outbound_messages

    def session_history(self) -> list[EnvelopeHistoryEntry] | None:
        """
        Get the message history associated with the context session.

        Returns:
            list[EnvelopeHistoryEntry] | None: The session history.
        """
        if self._message_history is None:
            log(self.logger, logging.ERROR, "No session history available")
            return None
        return self._message_history.get_session_messages(self._session)

    def get_agents_by_protocol(
        self,
        protocol_digest: str,
        limit: int = DEFAULT_SEARCH_LIMIT,
        logger: logging.Logger | None = None,
    ) -> list[str]:
        if not isinstance(protocol_digest, str) or not protocol_digest.startswith(
            "proto:"
        ):
            log(logger, logging.ERROR, f"Invalid protocol digest: {protocol_digest}")
            return []
        almanac_api_url = getattr(
            getattr(self._resolver, "_almanac_api_resolver", None),
            "_almanac_api_url",
            ALMANAC_API_URL,
        )
        response = requests.post(
            url=almanac_api_url + "/search",
            json={"text": protocol_digest[6:]},
            timeout=DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
        )
        if response.status_code == 200:
            data = response.json()
            agents = [agent["address"] for agent in data if agent["status"] == "active"]
            return agents[:limit]
        return []

    async def broadcast(
        self,
        destination_protocol: str,
        message: Model,
        limit: int = DEFAULT_SEARCH_LIMIT,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    ) -> list[MsgStatus]:
        agents = self.get_agents_by_protocol(
            destination_protocol, limit=limit, logger=self.logger
        )
        if not agents:
            log(
                self.logger,
                logging.ERROR,
                f"No active agents found for: {destination_protocol}",
            )
            return []

        if self.agent.address in agents:
            agents.remove(self.agent.address)
        futures = await asyncio.gather(
            *[
                self.send(
                    address,
                    message,
                    timeout=timeout,
                )
                for address in agents
            ]
        )
        log(self.logger, logging.DEBUG, f"Sent {len(futures)} messages")
        return futures

    def _is_valid_interval_message(self, schema_digest: str) -> bool:
        """
        Check if the message is a valid interval message.

        Args:
            schema_digest (str): The schema digest of the message to check.

        Returns:
            bool: Whether the message is a valid interval message.
        """
        if self._interval_messages:
            return schema_digest in self._interval_messages
        return True

    async def send(
        self,
        destination: str,
        message: Model,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    ) -> MsgStatus:
        """
        This is the pro-active send method which is used in on_event and
        on_interval methods. In these methods, interval messages are set but
        we don't have access properties that are only necessary in re-active
        contexts, like 'replies', 'message_received', or 'protocol'.
        """
        schema_digest: str = Model.build_schema_digest(message)
        message_body: JsonStr = message.model_dump_json()

        if not self._is_valid_interval_message(schema_digest):
            log(
                logger=self.logger,
                level=logging.ERROR,
                message=f"Invalid interval message: {message}",
            )
            return MsgStatus(
                status=DeliveryStatus.FAILED,
                detail="Invalid interval message",
                destination=destination,
                endpoint="",
                session=self._session,
            )

        return await self.send_raw(
            destination=destination,
            message_schema_digest=schema_digest,
            message_body=message_body,
            timeout=timeout,
        )

    async def send_raw(
        self,
        destination: str,
        message_schema_digest: str,
        message_body: JsonStr,
        sync: bool = False,
        wait_for_response: bool = False,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
        protocol_digest: str | None = None,
        queries: dict[str, asyncio.Future] | None = None,
    ) -> MsgStatus:
        # Extract address from destination agent identifier if present
        _, _, parsed_address = parse_identifier(destination)

        result = None
        if parsed_address:
            if sync or wait_for_response:
                dispatcher.register_pending_response(
                    sender=self.agent.address,
                    destination=parsed_address,
                    session=self._session,
                )
            # Handle local dispatch of messages
            if dispatcher.contains(parsed_address):
                result = await dispatch_local_message(
                    sender=self.agent.address,
                    destination=parsed_address,
                    schema_digest=message_schema_digest,
                    message=message_body,
                    session_id=self._session,
                )

            # Handle sync dispatch of messages
            elif queries and parsed_address in queries:
                queries[parsed_address].set_result(
                    (message_body, message_schema_digest)
                )
                del queries[parsed_address]
                result = MsgStatus(
                    status=DeliveryStatus.DELIVERED,
                    detail="Sync message resolved",
                    destination=parsed_address,
                    endpoint="",
                    session=self._session,
                )

            if parsed_address in self._outbound_messages:
                self._outbound_messages[parsed_address].append(
                    (message_body, message_schema_digest)
                )
            else:
                self._outbound_messages[parsed_address] = [
                    (message_body, message_schema_digest)
                ]

        if result is None:
            # Resolve destination using the resolver
            destination_address, endpoints = await self._resolver.resolve(destination)

            if not endpoints or not destination_address:
                log(
                    logger=self.logger,
                    level=logging.ERROR,
                    message=f"Unable to resolve destination endpoint for agent: {destination}",
                )
                result = MsgStatus(
                    status=DeliveryStatus.FAILED,
                    detail="Unable to resolve destination endpoint",
                    destination=destination,
                    endpoint="",
                    session=self._session,
                )
            else:
                # Calculate when the envelope expires
                expires = int(time()) + timeout

                # Handle external dispatch of messages
                env = Envelope(
                    version=1,
                    sender=self.agent.address,
                    target=destination_address,
                    session=self._session,
                    schema_digest=message_schema_digest,
                    protocol_digest=protocol_digest,
                    expires=expires,
                )
                env.encode_payload(message_body)
                env.sign(self.agent.identity)

                # Create awaitable future for MsgStatus and sync response
                fut = asyncio.Future()

                self._queue_envelope(env, endpoints, fut, sync)

                try:
                    result = await asyncio.wait_for(fut, timeout)
                except asyncio.TimeoutError:
                    log(
                        self.logger,
                        logging.ERROR,
                        "Timeout waiting for dispense response",
                    )
                    result = MsgStatus(
                        status=DeliveryStatus.FAILED,
                        detail="Timeout waiting for response",
                        destination=destination,
                        endpoint="",
                        session=self._session,
                    )

        if result.status == DeliveryStatus.DELIVERED and self._message_history:
            self._message_history.add_entry(
                EnvelopeHistoryEntry(
                    version=1,
                    sender=self.agent.address,
                    target=destination,
                    session=self._session,
                    schema_digest=message_schema_digest,
                    protocol_digest=protocol_digest,
                    payload=message_body,
                )
            )

        return result

    def _queue_envelope(
        self,
        envelope: Envelope,
        endpoints: list[str],
        response_future: asyncio.Future,
        sync: bool = False,
    ):
        """
        Queue an envelope for processing.

        Args:
            envelope (Envelope): The envelope to queue.
        """
        self._dispenser.add_envelope(envelope, endpoints, response_future, sync)

    async def send_and_receive(
        self,
        destination: str,
        message: Model,
        response_type: type[Model] | set[type[Model]],
        sync: bool = False,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    ) -> tuple[Model | None, MsgStatus]:
        """
        Send a message to the specified destination and receive a response.

        Args:
            destination (str): The destination address to send the message to.
            message (Model): The message to be sent.
            response_type (type[Model] | set[type[Model]]): The type(s) of the response message.
            sync (bool): Whether to send the message synchronously or asynchronously.
            timeout (int): The timeout for sending the message, in seconds.

        Returns:
            tuple[Model | None, MsgStatus]: The response message if received and delivery status
        """
        schema_digest = Model.build_schema_digest(message)

        msg_status: MsgStatus = await self.send_raw(
            destination=destination,
            message_schema_digest=schema_digest,
            message_body=message.model_dump_json(),
            sync=sync,
            wait_for_response=True,
            timeout=timeout,
        )

        _, _, parsed_address = parse_identifier(destination)

        if msg_status.status != DeliveryStatus.DELIVERED:
            dispatcher.cancel_pending_response(
                self.agent.address, parsed_address, self._session
            )
            return None, msg_status

        response_msg: MsgInfo | None = await dispatcher.wait_for_response(
            self.agent.address, parsed_address, self._session, timeout
        )

        if response_msg is None:
            log(self.logger, logging.ERROR, "Timeout waiting for response")
            return None, MsgStatus(
                status=DeliveryStatus.FAILED,
                detail="Timeout waiting for response",
                destination=destination,
                endpoint="",
                session=self._session,
            )

        response_types: set[type[Model]] = (
            {response_type} if isinstance(response_type, type) else response_type
        )

        response: Model | None = None
        for r_type in response_types:
            if response_msg.schema_digest == Model.build_schema_digest(r_type):
                try:
                    response = r_type.parse_raw(response_msg.message)
                    break
                except ValidationError:
                    pass

        if response is None:
            log(
                logger=self.logger,
                level=logging.ERROR,
                message=f"Received unexpected response: {response_msg}",
            )
            msg_status = MsgStatus(
                status=DeliveryStatus.FAILED,
                detail="Received unexpected response type",
                destination=destination,
                endpoint="",
                session=self._session,
            )

        return response, msg_status

    async def send_wallet_message(
        self,
        destination: str,
        text: str,
        msg_type: int = 1,
    ):  # TODO: restructure wallet messaging
        if self._wallet_messaging_client is not None:
            await self._wallet_messaging_client.send(destination, text, msg_type)
        else:
            log(
                logger=self.logger,
                level=logging.WARNING,
                message="Cannot send wallet message: no client available",
            )


class ExternalContext(InternalContext):
    """
    Represents the reactive context in which messages are handled and processed.

    Attributes:
        _message_received (MsgInfo): The received message.
        _queries (dict[str, asyncio.Future] | None): dictionary mapping query senders to their
            response Futures.
        _replies (dict[str, dict[str, type[Model]]] | None): Dictionary of allowed reply digests
            for each type of incoming message.
        _protocol (tuple[str, Protocol] | None): The supported protocol digest
            and the corresponding protocol.
    """

    def __init__(
        self,
        message_received: MsgInfo,
        queries: dict[str, asyncio.Future] | None = None,
        replies: dict[str, dict[str, type[Model]]] | None = None,
        protocol: tuple[str, "Protocol"] | None = None,
        **kwargs,
    ):
        """
        Initialize the ExternalContext instance and attributes needed from the InternalContext.

        Args:
            message_received (MsgInfo): Information about the received message.
            queries (dict[str, asyncio.Future]): Dictionary mapping query senders to their
                response Futures.
            replies (dict[str, dict[str, type[Model]]] | None): Dictionary of allowed replies
                for each type of incoming message.
            protocol (tuple[str, Protocol] | None): The optional tuple of protocols.
        """
        super().__init__(**kwargs)
        self._queries = queries or {}
        self._replies = replies
        self._message_received = message_received
        self._protocol = protocol or ("", None)

    def validate_replies(self, message_type: type[Model]) -> None:
        """
        If the context specifies replies, ensure that a valid reply was sent.

        Args:
            message_type (type[Model]): The type of the received message.
        """
        sender = self._message_received.sender
        received_digest = self._message_received.schema_digest

        # If no replies are defined, do not check for valid replies
        if not self._replies or received_digest not in self._replies:
            return

        # If replies are defined as empty, do not check for valid replies
        if self._replies[received_digest] == {}:
            return

        valid_replies = self._replies[received_digest] | {
            ERROR_MESSAGE_DIGEST: ErrorMessage
        }

        valid_reply_sent = False
        for target, msgs in self._outbound_messages.items():
            if target == sender and any(
                [digest in valid_replies for _, digest in msgs]
            ):
                valid_reply_sent = True
                break

        if not valid_reply_sent:
            valid_reply_types = {model.__name__ for model in valid_replies.values()}
            log(
                logger=self.logger,
                level=logging.ERROR,
                message=(
                    f"No valid reply {valid_reply_types} was sent to "
                    f"{sender} for received message: {message_type.__name__}."
                ),
            )

    async def send(
        self,
        destination: str,
        message: Model,
        timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    ) -> MsgStatus:
        """
        Send a message to the specified destination.

        Args:
            destination (str): The destination address to send the message to.
            message (Model): The message to be sent.
            timeout (int | None): The optional timeout for sending the message, in seconds.

        Returns:
            MsgStatus: The delivery status of the message.
        """
        schema_digest = Model.build_schema_digest(message)

        # This is the re-active send method
        # at this point we have received a message and have built a context
        # replies, message_received, and protocol are set

        return await self.send_raw(
            destination=destination,
            message_schema_digest=schema_digest,
            message_body=message.model_dump_json(),
            timeout=timeout,
            protocol_digest=self._protocol[0],
            queries=self._queries,
        )


ContextFactory = Callable[[], Context]



"""Endpoint Resolver."""

import logging
import random
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from enum import Enum

import aiohttp
from dateutil import parser
from pydantic import BaseModel
from uagents_core.helpers import weighted_random_sample
from uagents_core.identity import parse_identifier

from uagents.config import (
    ALMANAC_API_URL,
    DEFAULT_MAX_ENDPOINTS,
    MAINNET_PREFIX,
    TESTNET_PREFIX,
)
from uagents.network import (
    get_almanac_contract,
    get_name_service_contract,
)
from uagents.types import AgentNetwork
from uagents.utils import get_logger

LOGGER: logging.Logger = get_logger("resolver", logging.WARNING)


class AgentRecord(BaseModel):
    address: str
    weight: float


class DomainRecord(BaseModel):
    name: str
    agents: list[AgentRecord]


class DomainStatus(Enum):
    Registered = "Registered"
    Pending = "Pending"
    Checking = "Checking"
    Updating = "Updating"
    Deleting = "Deleting"
    Failed = "Failed"


class Domain(BaseModel):
    name: str
    status: DomainStatus
    expiry: datetime | None = None
    assigned_agents: list[AgentRecord]


def is_valid_prefix(prefix: str) -> bool:
    """
    Check if the given string is a valid prefix.

    Args:
        prefix (str): The prefix to be checked.

    Returns:
        bool: True if the prefix is valid; False otherwise.
    """
    valid_prefixes = [TESTNET_PREFIX, MAINNET_PREFIX, ""]
    return prefix in valid_prefixes


def parse_prefix(prefix: str) -> AgentNetwork:
    """
    Parse an agent prefix string into the corresponding network.

    Args:
        prefix (str): The prefix string to be parsed.

    Returns:
        AgentNetwork: The corresponding network.
    """
    if prefix in (TESTNET_PREFIX, ""):
        return "testnet"
    if prefix == MAINNET_PREFIX:
        return "mainnet"
    raise ValueError(f"Invalid prefix: {prefix}")


def query_record(agent_address: str, network: AgentNetwork) -> dict:
    """
    Query an agent record from the Almanac contract.

    Args:
        agent_address (str): The address of the agent.
        network (AgentNetwork): The network to query (mainnet or testnet).

    Returns:
        dict: The query result.
    """

    query_msg = {
        "query_record": {"agent_address": agent_address, "record_type": "service"}
    }

    contract = get_almanac_contract(network)
    if not contract:
        raise ValueError(f"Almanac contract not found for {network}.")

    return contract.query(query_msg)


def get_agent_address(name: str, network: AgentNetwork) -> str | None:
    """
    Get the agent address associated with the provided domain from the name service contract.

    Args:
        name (str): The name to query.
        network (AgentNetwork): The network to query (mainnet or testnet).

    Returns:
        str | None: The associated agent address if found.
    """
    query_msg = {"query_domain_record": {"domain": f"{name}"}}
    result = get_name_service_contract(network).query(query_msg)
    if result["record"] is not None:
        registered_records = result["record"]["records"][0]["agent_address"]["records"]
        if len(registered_records) > 0:
            addresses = [val.get("address") for val in registered_records]
            weights = [val.get("weight") for val in registered_records]
            selected_address_list = weighted_random_sample(addresses, weights=weights)
            selected_address = (
                selected_address_list[0] if selected_address_list else None
            )
            return selected_address
    return None


def build_identifier(
    prefix: str | None = None, name: str | None = None, address: str | None = None
) -> str:
    """
    Build an agent identifier string from prefix, name, and address.

    Args:
        prefix (str): The prefix to be used in the identifier.
        name (str): The name to be used in the identifier.
        address (str): The address to be used in the identifier.

    Returns:
        str: The constructed identifier string.
    """

    identifier = ""
    if prefix:
        identifier += f"{prefix}://"
    if name:
        identifier += name + ("/" if address else "")
    if address:
        identifier += address

    return identifier


class Resolver(ABC):
    @abstractmethod
    async def resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination to an address and endpoint.

        Args:
            destination (str): The destination name or address to resolve.

        Returns:
            tuple[str | None, list[str]]: The address (if available) and resolved endpoints.
        """
        raise NotImplementedError


class GlobalResolver(Resolver):
    def __init__(
        self, max_endpoints: int | None = None, almanac_api_url: str | None = None
    ):
        """
        Initialize the GlobalResolver.

        Args:
            max_endpoints (int | None): The maximum number of endpoints to return.
            almanac_api_url (str | None): The url for almanac api
        """
        self._max_endpoints = max_endpoints or DEFAULT_MAX_ENDPOINTS
        self._almanac_api_resolver = AlmanacApiResolver(
            max_endpoints=self._max_endpoints, almanac_api_url=almanac_api_url
        )
        self._name_service_resolver = NameServiceResolver(
            max_endpoints=self._max_endpoints, almanac_api_url=almanac_api_url
        )

    async def resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination using the appropriate resolver.

        Args:
            destination (str): The destination name or address to resolve.

        Returns:
            tuple[str | None, list[str]]: The address (if available) and resolved endpoints.
        """

        prefix, _, address = parse_identifier(destination)

        if is_valid_prefix(prefix):
            resolver = (
                self._almanac_api_resolver if address else self._name_service_resolver
            )
            return await resolver.resolve(destination)

        return None, []


class AlmanacContractResolver(Resolver):
    def __init__(self, max_endpoints: int | None = None):
        """
        Initialize the AlmanacContractResolver.

        Args:
            max_endpoints (int | None): The maximum number of endpoints to return.
        """
        self._max_endpoints = max_endpoints or DEFAULT_MAX_ENDPOINTS

    async def resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination using the Almanac contract.

        Args:
            destination (str): The destination address to resolve.

        Returns:
            tuple[str | None, list[str]]: The address and resolved endpoints.
        """
        prefix, _, address = parse_identifier(destination)

        result: dict | None = None
        if prefix == MAINNET_PREFIX:
            result = query_record(agent_address=address, network="mainnet")
        elif prefix == TESTNET_PREFIX:
            result = query_record(agent_address=address, network="testnet")
        elif prefix == "":
            for network in ["mainnet", "testnet"]:
                try:
                    result = query_record(agent_address=address, network="mainnet")
                    if result is not None:
                        break
                except ValueError:
                    if network == "testnet":
                        raise

        if result is not None:
            record = result.get("record") or {}
            endpoint_list = (
                record.get("record", {}).get("service", {}).get("endpoints", [])
            )

            if len(endpoint_list) > 0:
                endpoints = [val.get("url") for val in endpoint_list]
                weights = [val.get("weight") for val in endpoint_list]
                return address, weighted_random_sample(
                    items=endpoints,
                    weights=weights,
                    k=min(self._max_endpoints, len(endpoints)),
                )

        return None, []


class AlmanacApiResolver(Resolver):
    def __init__(
        self, max_endpoints: int | None = None, almanac_api_url: str | None = None
    ):
        """
        Initialize the AlmanacApiResolver.

        Args:
            max_endpoints (int | None): The maximum number of endpoints to return.
            almanac_api_url (str | None): The url for almanac api
        """
        self._max_endpoints = max_endpoints or DEFAULT_MAX_ENDPOINTS
        self._almanac_api_url = almanac_api_url or ALMANAC_API_URL
        self._almanac_contract_resolver = AlmanacContractResolver(
            max_endpoints=self._max_endpoints
        )

    async def _api_resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination using the Almanac API.

        Args:
            destination (str): The destination address to resolve.

        Returns:
            tuple[str | None, list[str]]: The address and resolved endpoints.
        """
        try:
            prefix, _, address = parse_identifier(destination)

            params = {"prefix": prefix} if prefix else None
            async with (
                aiohttp.ClientSession() as session,
                session.get(
                    url=f"{self._almanac_api_url}/agents/{address}",
                    params=params,
                    timeout=aiohttp.ClientTimeout(total=5),
                ) as response,
            ):
                if response.status != 200:
                    LOGGER.debug(
                        f"Failed to resolve agent {address} from {self._almanac_api_url}, "
                        "resolving via Almanac contract..."
                    )
                    return None, []

                agent = await response.json()

            expiry_str = agent.get("expiry", None)
            if expiry_str is None:
                return None, []

            expiry = parser.parse(expiry_str)
            current_time = datetime.now(timezone.utc)
            endpoint_list = agent.get("endpoints", [])

            if len(endpoint_list) > 0 and expiry > current_time:
                endpoints = [val.get("url") for val in endpoint_list]
                weights = [val.get("weight") for val in endpoint_list]
                return address, weighted_random_sample(
                    endpoints,
                    weights=weights,
                    k=min(self._max_endpoints, len(endpoints)),
                )
        except Exception as e:
            LOGGER.error(
                f"Error in AlmanacApiResolver when resolving {destination}: {e}"
            )

        return None, []

    async def resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination using the Almanac API.
        If the resolution using API fails, it retries using the Almanac Contract.

        Args:
            destination (str): The destination address to resolve.

        Returns:
            tuple[str | None, list[str]]: The address and resolved endpoints.
        """
        address, endpoints = await self._api_resolve(destination)
        return (
            (address, endpoints)
            if address is not None
            else await self._almanac_contract_resolver.resolve(destination)
        )


class NameServiceResolver(Resolver):
    def __init__(
        self, max_endpoints: int | None = None, almanac_api_url: str | None = None
    ):
        """
        Initialize the NameServiceResolver.

        Args:
            max_endpoints (int | None): The maximum number of endpoints to return.
        """
        self._max_endpoints = max_endpoints or DEFAULT_MAX_ENDPOINTS
        self._almanac_api_url = almanac_api_url
        self._almanac_api_resolver = AlmanacApiResolver(
            almanac_api_url=almanac_api_url, max_endpoints=self._max_endpoints
        )

    async def _api_resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination using the Almanac Domains API.

        Args:
            destination (str): The agent identifier to resolve.

        Returns:
            Tuple[Optional[str], List[str]]: The address (if available) and resolved endpoints.
        """
        try:
            prefix, domain, _ = parse_identifier(destination)

            params = {"prefix": prefix} if prefix else None
            async with (
                aiohttp.ClientSession() as session,
                session.get(
                    f"{self._almanac_api_url}/domains/{domain}",
                    params=params,
                    timeout=aiohttp.ClientTimeout(total=5),
                ) as response,
            ):
                if response.status != 200:
                    LOGGER.debug(
                        f"Failed to resolve name {domain} from {self._almanac_api_url}: "
                        f"{response.status}: {await response.text()}"
                    )
                    return None, []

                domain_record = Domain.model_validate(await response.json())

            agent_records = domain_record.assigned_agents
            if len(agent_records) == 0:
                return None, []
            elif len(agent_records) == 1:
                address = agent_records[0].address
            else:
                addresses = [val.address for val in agent_records]
                weights = [val.weight for val in agent_records]
                address = weighted_random_sample(addresses, weights=weights, k=1)[0]

            identifier = build_identifier(prefix=prefix, address=address)
            return await self._almanac_api_resolver.resolve(identifier)

        except Exception as ex:
            LOGGER.error(f"Error when resolving {destination}: {ex}")
            return None, []

    async def resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination using the NameService contract.

        Args:
            destination (str): The destination name to resolve.

        Returns:
            tuple[str | None, list[str]]: The address (if available) and resolved endpoints.
        """
        prefix, name, _ = parse_identifier(destination)

        api_result = await self._api_resolve(destination)

        if api_result:
            return api_result

        address: str | None = None
        if prefix == MAINNET_PREFIX:
            address = get_agent_address(name, "mainnet")
        elif prefix == TESTNET_PREFIX:
            address = get_agent_address(name, "testnet")
        elif prefix == "":
            for network in ["mainnet", "testnet"]:
                address = get_agent_address(name, network)  # type: ignore
                if address is not None:
                    break
        else:
            raise ValueError(f"Invalid prefix: {prefix}")

        if address is not None:
            identifier = build_identifier(prefix=prefix, address=address)
            return await self._almanac_api_resolver.resolve(identifier)

        return None, []


class RulesBasedResolver(Resolver):
    def __init__(self, rules: dict[str, str], max_endpoints: int | None = None):
        """
        Initialize the RulesBasedResolver with the provided rules.

        Args:
            rules (dict[str, str]): A dictionary of rules mapping destinations to endpoints.
            max_endpoints (int | None): The maximum number of endpoints to return.
        """
        self._rules = rules
        self._max_endpoints = max_endpoints or DEFAULT_MAX_ENDPOINTS

    async def resolve(self, destination: str) -> tuple[str | None, list[str]]:
        """
        Resolve the destination using the provided rules.

        Args:
            destination (str): The destination to resolve.

        Returns:
            tuple[str | None, list[str]]: The address and resolved endpoints.
        """
        endpoints = self._rules.get(destination)
        if isinstance(endpoints, str):
            endpoints = [endpoints]
        elif endpoints is None:
            endpoints = []
        if len(endpoints) > self._max_endpoints:
            endpoints = random.sample(
                population=endpoints, k=min(self._max_endpoints, len(endpoints))
            )
        return destination, endpoints



import base64
import hashlib
import json

from uagents_core.identity import Identity, encode_length_prefixed


def sign_registration(
    identity: Identity,
    contract_address: str,
    timestamp: int,
    wallet_address: str,
) -> str:
    """Sign the registration data for the Almanac contract."""
    hasher = hashlib.sha256()
    hasher.update(encode_length_prefixed(contract_address))
    hasher.update(encode_length_prefixed(identity.address))
    hasher.update(encode_length_prefixed(timestamp))
    hasher.update(encode_length_prefixed(wallet_address))
    return identity.sign_digest(hasher.digest())


def sign_arbitrary(identity: Identity, data: bytes) -> tuple[str, str]:
    """
    Sign arbitrary data with the given identity.

    Only used for wallet messaging.
    """
    sign_doc = {
        "chain_id": "",
        "account_number": "0",
        "sequence": "0",
        "fee": {
            "gas": "0",
            "amount": [],
        },
        "msgs": [
            {
                "type": "sign/MsgSignData",
                "value": {
                    "signer": identity.address,
                    "data": base64.b64encode(data).decode(),
                },
            },
        ],
        "memo": "",
    }

    raw_sign_doc = json.dumps(
        sign_doc,
        sort_keys=True,
        separators=(",", ":"),
    ).encode()
    signature = identity.sign_b64(raw_sign_doc)
    enc_sign_doc = base64.b64encode(raw_sign_doc).decode()

    return enc_sign_doc, signature



from datetime import datetime
from typing import Annotated, Literal

import requests
from pydantic import BaseModel, Field

from uagents.config import AGENTVERSE_URL

SEARCH_API_URL = AGENTVERSE_URL + "/v1/search/agents"

StatusType = Literal["active", "inactive"]
AgentType = Literal["hosted", "local", "mailbox", "proxy", "custom"]
AgentCategory = Literal["fetch-ai", "verified", "community"]


class ProtocolRepresentation(BaseModel):
    # the name of the protocol
    name: str

    # the version of the protocol
    version: str

    # the digest of the protocol
    digest: str


class AgentGeoLocation(BaseModel):
    # the latitude of the agent
    latitude: float

    # the longitude of the agent
    longitude: float

    # the radius in meters defining the area of effect of the agent
    radius: float = 0


class Agent(BaseModel):
    # the address of the agent (this should be used as the id of the agent)
    address: str

    # the public name of the agent
    name: str

    # the contents of the readme file
    readme: str

    # the list of protocols supported by the agent
    protocols: list[ProtocolRepresentation]

    # the href for the avatar image for the agent
    avatar_href: str | None

    # the total interactions for this agent
    total_interactions: int

    # the number of interactions in the last 90 days
    recent_interactions: int

    # agent rating a number between 0 and 5
    rating: float | None

    # the status of the agent
    status: StatusType

    # the type of agent
    type: AgentType

    # the creator of the agent
    category: AgentCategory

    # signaled if the agent is featured or not
    featured: bool = False

    # the geolocation of the agent
    geo_location: AgentGeoLocation | None

    # the time at which the agent was last updated at
    last_updated: datetime

    # the time at which the agent was first visible or created
    created_at: datetime


class AgentFilters(BaseModel):
    """The set of filters that should be applied to the agent search entries"""

    # The state of the agent, i.e. is it alive or not
    state: list[StatusType] = Field(default_factory=list)

    # The category of the creator of the agent
    category: list[AgentCategory] = Field(default_factory=list)

    # The category of how the agent is hosted
    agent_type: list[AgentType] = Field(default_factory=list)


SortType = Literal["relevancy", "created-at", "last-modified", "interactions"]
SortDirection = Literal["asc", "desc"]


class AgentSearchCriteria(BaseModel):
    """The search criteria that can be set for the agent search"""

    # The set of filters that should be applied to the search
    filters: AgentFilters = AgentFilters()

    # The type of sorting that should be applied to the search results, relevancy is the default
    sort: SortType = "relevancy"

    # The direction of the sorting, ascending or descending
    direction: SortDirection = "asc"

    # The optional search text that should be included. This should not be a filter mechanism but
    # entries that are closer to the search text should be ranked higher
    search_text: str | None = None

    # The offset of the search results for pagination
    offset: int = 0

    # The limit of the search results for pagination
    limit: int = 30


class AgentGeoFilter(BaseModel):
    """The geo filter that can be applied to the agent search"""

    # The latitude of the location
    latitude: Annotated[float, Field(ge=-90, le=90)]

    # The longitude of the location
    longitude: Annotated[float, Field(ge=-180, le=180)]

    # The radius of the search in meters
    radius: Annotated[float, Field(gt=0)]


class AgentGeoSearchCriteria(AgentSearchCriteria):
    """The search criteria that can be set for the agent search"""

    # The geo filter that can be applied to the search
    geo_filter: AgentGeoFilter


def _geosearch_agents(criteria: AgentGeoSearchCriteria) -> list[Agent]:
    # NOTE: currently results will be returned based on radius overlap, i.e., results can
    # include agents that are farther away then the specified radius.

    # filter only for active agents
    criteria.filters = AgentFilters(state=["active"])
    response = requests.post(
        url=SEARCH_API_URL + "/geo",
        json=criteria.model_dump(),
        timeout=5,
    )
    if response.status_code == 200:
        data = response.json()
        agents = [Agent.model_validate(agent) for agent in data["agents"]]
        return agents
    return []


def _search_agents(criteria: AgentSearchCriteria) -> list[Agent]:
    response = requests.post(
        url=SEARCH_API_URL,
        json=criteria.model_dump(),
        timeout=5,
    )
    if response.status_code == 200:
        data = response.json()
        agents = [Agent.model_validate(agent) for agent in data["agents"]]
        return agents
    return []


def geosearch_agents_by_proximity(
    latitude: float,
    longitude: float,
    radius: float,
    limit: int = 30,
) -> list[Agent]:
    """
    Return all agents in a circle around the given coordinates that match the given search criteria
    """
    criteria = AgentGeoSearchCriteria(
        geo_filter=AgentGeoFilter(
            latitude=latitude, longitude=longitude, radius=radius
        ),
        limit=limit,
    )
    return _geosearch_agents(criteria)


def geosearch_agents_by_protocol(
    latitude: float,
    longitude: float,
    radius: float,
    protocol_digest: str,
    limit: int = 30,
) -> list[Agent]:
    """
    Return all agents in a circle around the given coordinates that match the given search criteria
    """
    criteria = AgentGeoSearchCriteria(
        geo_filter=AgentGeoFilter(
            latitude=latitude, longitude=longitude, radius=radius
        ),
        limit=limit,
    )
    unfiltered_geoagents = _geosearch_agents(criteria)
    filtered_agents = [
        agent
        for agent in unfiltered_geoagents
        if protocol_digest in [protocol.digest for protocol in agent.protocols]
    ]
    return filtered_agents


def geosearch_agents_by_text(
    latitude: float, longitude: float, radius: float, search_text: str, limit: int = 30
) -> list[Agent]:
    """
    Return all agents in a circle around the given coordinates that match the given search_text
    """
    criteria = AgentGeoSearchCriteria(
        geo_filter=AgentGeoFilter(
            latitude=latitude, longitude=longitude, radius=radius
        ),
        limit=limit,
        search_text=search_text,
    )
    return _geosearch_agents(criteria)


def search_agents_by_protocol(protocol_digest: str, limit: int = 30) -> list[Agent]:
    """Return all agents that match the given search criteria"""
    criteria = AgentSearchCriteria(
        filters=AgentFilters(state=["active"]),
        search_text=protocol_digest,
        limit=limit,
    )
    unfiltered_geoagents = _search_agents(criteria)
    filtered_agents = [
        agent
        for agent in unfiltered_geoagents
        if protocol_digest in [protocol.digest for protocol in agent.protocols]
    ]
    return filtered_agents


def search_agents_by_text(search_text: str, limit: int = 30) -> list[Agent]:
    """Return all agents that match the given search_text"""
    criteria = AgentSearchCriteria(
        filters=AgentFilters(state=["active"]),
        search_text=search_text,
        limit=limit,
    )
    return _search_agents(criteria)



import logging
import sys

from uvicorn.logging import DefaultFormatter

logging.basicConfig(level=logging.INFO)


def get_logger(logger_name: str, level: int | str = logging.INFO):
    """Get a logger with the given name using uvicorn's default formatter."""
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    log_handler = logging.StreamHandler(sys.stdout)
    log_handler.setFormatter(
        DefaultFormatter(fmt="%(levelprefix)s [%(name)5s]: %(message)s")
    )
    logger.addHandler(log_handler)
    logger.propagate = False
    return logger


def log(logger: logging.Logger | None, level: int, message: str):
    """
    Log a message with the given logger and level.

    Args:
        logger (logging.Logger | None): The logger to use.
        level (int): The logging level.
        message (str): The message to log.
    """
    if logger:
        logger.log(level, message)
    else:
        BACKUP_LOGGER.log(level, message)


def set_global_log_level(level: int | str):
    """
    Set the log level for all modules globally. Can still be overruled manually.

    Args:
        level (int | str): The logging level as defined in _logging_.
    """
    logging.basicConfig(level=level)
    for name in logging.Logger.manager.loggerDict:
        logging.getLogger(name).setLevel(level)


BACKUP_LOGGER: logging.Logger = get_logger("uagents", logging.INFO)



"""Agent's Setup."""

from logging import Logger

from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.faucet import FaucetApi
from cosmpy.crypto.address import Address

from uagents.config import TESTNET_REGISTRATION_FEE
from uagents.network import get_faucet, get_ledger
from uagents.utils import get_logger

LOGGER: Logger = get_logger("setup")


def fund_agent_if_low(
    wallet_address: str, min_balance: int = TESTNET_REGISTRATION_FEE
) -> None:
    """
    Checks the agent's wallet balance and adds testnet funds if it's below min_balance.

    Args:
        wallet_address (str): The wallet address of the agent.
        min_balance (int): The minimum balance required. Defaults to TESTNET_REGISTRATION_FEE.
    """
    ledger: LedgerClient = get_ledger()
    faucet: FaucetApi = get_faucet()

    agent_balance = ledger.query_bank_balance(Address(wallet_address))

    if agent_balance < min_balance:
        try:
            LOGGER.info("Adding testnet funds to agent...")
            faucet.get_wealth(wallet_address)
            LOGGER.info("Adding testnet funds to agent...complete")
        except Exception as ex:
            LOGGER.error(f"Failed to add testnet funds to agent: {str(ex)}")



"""Query Envelopes."""

from typing_extensions import deprecated
from uagents_core.envelope import Envelope
from uagents_core.identity import generate_user_address
from uagents_core.models import Model
from uagents_core.types import MsgStatus

from uagents.communication import send_sync_message
from uagents.resolver import Resolver


@deprecated(
    "Query is deprecated and will be removed in a future release, use send_sync_message instead."
)
async def query(
    destination: str,
    message: Model,
    resolver: Resolver | None = None,
    timeout: int = 30,
) -> MsgStatus | Envelope:
    """
    Query a remote agent with a message and retrieve the response envelope.

    Args:
        destination (str): The destination address of the remote agent.
        message (Model): The message to send.
        resolver (Resolver | None): The resolver to use for endpoint resolution.
        Defaults to GlobalResolver.
        timeout (int): The timeout for the query in seconds. Defaults to 30.

    Returns:
        MsgStatus | Envelope: The response envelope if successful, otherwise MsgStatus.
    """
    response = await send_sync_message(
        destination=destination,
        message=message,
        response_type=None,
        sender=generate_user_address(),
        resolver=resolver,
        timeout=timeout,
    )
    if isinstance(response, (MsgStatus, Envelope)):
        return response
    raise ValueError("Invalid response received.")



from uagents_core.models import Field, Model
from uagents_core.protocol import ProtocolSpecification

from .agent import Agent, Bureau
from .context import Context
from .protocol import Protocol

__all__ = [
    "Agent",
    "Bureau",
    "Context",
    "Field",
    "Model",
    "Protocol",
    "ProtocolSpecification",
]



"""Exchange Protocol"""

import functools
from collections.abc import Awaitable, Callable
from logging import Logger
from typing import Any

from typing_extensions import deprecated
from uagents_core.models import Model
from uagents_core.protocol import ProtocolSpecification

from uagents.config import get_logger
from uagents.types import IntervalCallback, MessageCallback

logger: Logger = get_logger("protocol")


class Protocol:
    """
    The Protocol class encapsulates a particular set of functionalities for an agent.
    It typically relates to the exchange of messages between agents for executing some task.
    It includes the message (model) types it supports, the allowed replies, and the
    interval message handlers that define the logic of the protocol.
    """

    def __init__(
        self,
        name: str | None = None,
        version: str | None = None,
        spec: ProtocolSpecification | None = None,
        role: str | None = None,
    ) -> None:
        """
        Initialize a Protocol instance.

        Args:
            name (str | None): The name of the protocol. Defaults to None.
            version (str | None): The version of the protocol. Defaults to None.
            spec (ProtocolSpecification | None): The protocol specification. Defaults to None.
            role (str | None): The role that the protocol will implement. Defaults to None.
        """
        self._interval_handlers: list[tuple[IntervalCallback, float]] = []
        self._interval_messages: set[str] = set()
        self._signed_message_handlers: dict[str, MessageCallback] = {}
        self._unsigned_message_handlers: dict[str, MessageCallback] = {}
        self._models: dict[str, type[Model]] = {}
        self._replies: dict[str, dict[str, type[Model]]] = {}
        self._digest = ""
        self._role: str | None = None
        self._locked = False

        if spec:
            self._init_from_spec(spec, role, name, version)
        else:
            self._spec = ProtocolSpecification(
                name=name or "", version=version or "0.1.0", interactions={}
            )

    @property
    def intervals(self) -> list[tuple[IntervalCallback, float]]:
        """
        Property to access the interval handlers.

        Returns:
            list[tuple[IntervalCallback, float]]: List of interval handlers and their periods.
        """
        return self._interval_handlers

    @property
    def models(self) -> dict[str, type[Model]]:
        """
        Property to access the registered models.

        Returns:
            dict[str, type[Model]]: Dictionary of registered models with schema digests as keys.
        """
        return self._models

    @property
    def replies(self) -> dict[str, dict[str, type[Model]]]:
        """
        Property to access the registered replies.

        Returns:
            dict[str, dict[str, type[Model]]]: Dictionary mapping message schema digests to their
            allowed replies.
        """
        return self._replies

    @property
    def interval_messages(self) -> set[str]:
        """
        Property to access the interval message digests.

        Returns:
            set[str]: Set of message digests that may be sent by interval handlers.
        """
        return self._interval_messages

    @property
    def signed_message_handlers(self) -> dict[str, MessageCallback]:
        """
        Property to access the signed message handlers.

        Returns:
            dict[str, MessageCallback]: Dictionary mapping message schema digests to their handlers.
        """
        return self._signed_message_handlers

    @property
    def unsigned_message_handlers(self) -> dict[str, MessageCallback]:
        """
        Property to access the unsigned message handlers.

        Returns:
            dict[str, MessageCallback]: Dictionary mapping message schema digests to their handlers.
        """
        return self._unsigned_message_handlers

    @property
    def name(self) -> str:
        """
        Property to access the protocol name.

        Returns:
            str: The protocol name.
        """
        return self._spec.name

    @property
    def version(self) -> str:
        """
        Property to access the protocol version.

        Returns:
            str: The protocol version.
        """
        return self._spec.version

    @property
    def canonical_name(self) -> str:
        """
        Property to access the canonical name of the protocol ('name:version').

        Returns:
            str: The canonical name of the protocol.
        """
        return f"{self.name}:{self.version}"

    @property
    def digest(self) -> str:
        """
        Property to access the digest of the protocol's manifest.

        Returns:
            str: The digest of the protocol's manifest.
        """
        return self.manifest()["metadata"]["digest"]

    @property
    def spec(self) -> ProtocolSpecification:
        """
        Property to access the protocol specification.

        Returns:
            ProtocolSpecification: The protocol specification.
        """
        return self._spec

    def _init_from_spec(
        self,
        spec: ProtocolSpecification,
        role: str | None,
        name: str | None,
        version: str | None,
    ) -> None:
        """
        Initialize the protocol interactions from a specification.

        Args:
            spec (ProtocolSpecification): The protocol specification.
            role (str | None): The role that the protocol will implement.
            name (str | None): The name of the protocol.
            version (str | None): The version of the protocol.
        """
        if name and spec.name and name != spec.name:
            logger.warning(
                f"Protocol specification name '{spec.name}' overrides given protocol name '{name}'"
            )
        if version and spec.version and version != spec.version:
            logger.warning(
                f"Protocol specification version '{spec.version}' "
                f"overrides given protocol version '{version}'"
            )
        spec.name = spec.name or name or ""
        spec.version = spec.version or version or "0.1.0"

        if spec.roles:
            if role is None:
                raise ValueError("Role must be specified for a protocol with roles")
            if role not in spec.roles:
                raise ValueError(f"Role '{role}' not found in protocol roles")
            self._role = role
            interactions = {
                model: replies
                for model, replies in spec.interactions.items()
                if model in spec.roles[role]
            }
        else:
            interactions = spec.interactions

        for model, replies in interactions.items():
            self._add_interaction(model, replies)

        self._spec = spec
        self._locked = True

    def _add_interaction(
        self, model: type[Model], replies: set[type[Model]] | None = None
    ) -> None:
        """
        Add a message model to the protocol along with replies if specified.

        Args:
            model (type[Model]): The message model type.
            replies (set[type[Model]] | None): The associated reply types. Defaults to None.
        """
        model_digest = Model.build_schema_digest(model)
        self._models[model_digest] = model

        if replies is not None:
            self.replies[model_digest] = {
                Model.build_schema_digest(reply): reply for reply in replies
            }

    def on_interval(
        self,
        period: float,
        messages: type[Model] | set[type[Model]] | None = None,
    ) -> Callable:
        """
        Decorator to register an interval handler for the protocol.

        Args:
            period (float): The interval period in seconds.
            messages (type[Model] | set[type[Model]] | None): The associated message types.

        Returns:
            Callable: The decorator to register the interval handler.
        """

        def decorator_on_interval(func: IntervalCallback):
            @functools.wraps(func)
            def handler(*args, **kwargs) -> Awaitable[None]:
                return func(*args, **kwargs)

            self._add_interval_handler(period, func, messages)

            return handler

        return decorator_on_interval

    def _add_interval_handler(
        self,
        period: float,
        func: IntervalCallback,
        messages: type[Model] | set[type[Model]] | None,
    ) -> None:
        """
        Add an interval handler to the protocol.

        Args:
            period (float): The interval period in seconds.
            func (IntervalCallback): The interval handler function.
            messages (type[Model] | set[type[Model]] | None): The associated message types.
        """
        # store the interval handler for later
        self._interval_handlers.append((func, period))

        # if message types are specified, store these for validation
        if messages is not None:
            if not isinstance(messages, set):
                messages = {messages}
            for message in messages:
                message_digest = Model.build_schema_digest(message)
                self._interval_messages.add(message_digest)

    @deprecated(
        "on_query is deprecated and will be removed in a future release, use on_rest instead."
    )
    def on_query(
        self,
        model: type[Model],
        replies: type[Model] | set[type[Model]] | None = None,
    ) -> Callable:
        """
        Decorator to register a query handler for the protocol.

        Args:
            model (type[Model]): The message model type.
            replies (type[Model] | set[type[Model]] | None): The associated reply types.

        Returns:
            Callable: The decorator to register the query handler.
        """
        return self.on_message(model, replies, allow_unverified=True)

    def on_message(
        self,
        model: type[Model],
        replies: type[Model] | set[type[Model]] | None = None,
        allow_unverified: bool = False,
    ) -> Callable:
        """
        Decorator to register a message handler for the protocol.

        Args:
            model (type[Model]): The message model type.
            replies (type[Model] | set[type[Model]] | None): The associated reply types.
            allow_unverified (bool, optional): Whether to allow unverified messages.

        Returns:
            Callable: The decorator to register the message handler.
        """
        if self._locked:
            if model not in self._models.values():
                raise ValueError(
                    f"Cannot add interaction '{model.__name__}' "
                    f"to locked protocol {self.canonical_name}"
                )
            model_digest = Model.build_schema_digest(model)
            if replies is not None and replies != self._replies.get(model_digest):
                raise ValueError(
                    f"Specified replies for '{model.__name__}' ({replies}) "
                    f"do not match the replies defined in the protocol specification "
                    f"for {self.canonical_name}: {self._replies.get(model_digest)}"
                )

        def decorator_on_message(func: MessageCallback):
            @functools.wraps(func)
            def handler(*args, **kwargs):
                return func(*args, **kwargs)

            self._add_message_handler(model, func, replies, allow_unverified)

            return handler

        return decorator_on_message

    def _add_message_handler(
        self,
        model: type[Model],
        func: MessageCallback,
        replies: type[Model] | set[type[Model]] | None,
        allow_unverified: bool = False,
    ):
        """
        Add a message handler to the protocol.

        Args:
            model (type[Model]): The message model type.
            func (MessageCallback): The message handler function.
            replies (type[Model] | set[type[Model]] | None): The associated reply types.
            allow_unverified (bool, optional): Whether to allow unverified messages.
        """
        model_digest = Model.build_schema_digest(model)

        # update the model database
        self._models[model_digest] = model
        if allow_unverified:
            self._unsigned_message_handlers[model_digest] = func
        else:
            self._signed_message_handlers[model_digest] = func
        if replies is None:
            replies = set()
        if not isinstance(replies, set):
            replies = {replies}
        if not self._locked:
            self._spec.interactions[model] = replies
        self._replies[model_digest] = {
            Model.build_schema_digest(reply): reply for reply in replies
        }

    def manifest(self) -> dict[str, Any]:
        """
        Generate the protocol's manifest, a long-form machine readable description of the
        protocol details and interface.

        Returns:
            dict[str, Any]: The protocol's manifest.
        """
        return self._spec.manifest(role=self._role)

    def verify(self) -> bool:
        """
        Check if the protocol implements all interactions of its specification.

        Returns:
            bool: True if the protocol implements the role, False otherwise.
        """
        if not self._locked:
            # No specification to verify against
            return True

        message_handlers: dict[str, MessageCallback] = (
            self._signed_message_handlers | self._unsigned_message_handlers
        )

        result = True
        # verify that all models required by the role are implemented
        for digest, model in self._models.items():
            if digest not in message_handlers:
                logger.error(
                    f"Protocol {self.canonical_name} does not implement "
                    f"a handler for the model '{model.__name__}'"
                    + f" required for the role '{self._role}'"
                    if self._role
                    else ""
                )
                result = False

        return result

    @staticmethod
    def compute_digest(manifest: dict[str, Any]) -> str:
        """
        Compute the digest of a given manifest.

        Args:
            manifest (dict[str, Any]): The manifest to compute the digest for.

        Returns:
            str: The computed digest.
        """
        return ProtocolSpecification.compute_digest(manifest)



"""Dialogue class aka. blueprint for protocols."""

import functools
import warnings
from collections.abc import Awaitable, Callable
from datetime import datetime, timedelta
from typing import Any
from uuid import UUID

from uagents_core.models import ErrorMessage
from uagents_core.types import DeliveryStatus, JsonStr, MsgStatus

from uagents import Context, Model, Protocol
from uagents.storage import KeyValueStore, StorageAPI

DEFAULT_SESSION_TIMEOUT_IN_SECONDS = 60
TARGET_UUID_VERSION = 4

MessageCallback = Callable[["Context", str, Any], Awaitable[None]]


class Node:
    """A node represents a state in the dialogue."""

    def __init__(
        self,
        name: str,
        description: str,
        initial: bool = False,
    ) -> None:
        self.name = name
        self.description = description
        self.initial = initial
        self.final = False


class Edge:
    """An edge represents a transition between two states in the dialogue."""

    def __init__(
        self,
        name: str,
        description: str,
        parent: Node | None,  # tail
        child: Node,  # head
    ) -> None:
        self.name = name
        self.description = description
        self.parent = parent
        self.child = child
        self.starter: bool = False
        self.ender: bool = False
        self._model: type[Model] | None = None
        self._func: MessageCallback | None = None
        self._efunc: MessageCallback | None = None

    @property
    def model(self) -> type[Model] | None:
        """The message model type that is associated with the edge."""
        return self._model

    @model.setter
    def model(self, model: type[Model]) -> None:
        """Set the message model type for the edge."""
        self._model = model

    @property
    def func(self) -> MessageCallback | None:
        """The message handler that is associated with the edge."""
        return self._func

    @func.setter
    def func(self, func: MessageCallback) -> None:
        """Set the message handler that will be called when a message is received."""
        self._func = func

    @property
    def efunc(self) -> MessageCallback | None:
        """The edge handler that is associated with the edge."""
        return self._efunc

    def set_edge_handler(self, model: type[Model], func: MessageCallback) -> None:
        """
        Set the edge handler that will be called when a message is received
        This handler can not be overwritten by a decorator.
        """
        if self._model and self._model is not model:
            raise ValueError("Functionality already set with a different model!")
        self._model = model
        self._efunc = func

    def set_message_handler(self, model: type[Model], func: MessageCallback) -> None:
        """
        Set the default message handler for the edge that will be overwritten if
        a decorator defines a new function to be called.
        """
        if self._model and self._model is not model:
            raise ValueError("Functionality already set with a different model!")
        self._model = model
        self._func = func


class Dialogue(Protocol):
    """
    A dialogue is a protocol with added functionality to handle the enforcement
    of a sequence of messages.
    The instance of this class is the local representation of the dialogue,
    i.e. the definition of a pattern of messages that are exchanged between
    two participants.

    When defining a pattern of Nodes and Edges, the dialogue will automatically
    be validated for cycles and the rules will be derived from the graph.

    The only thing left to do is to add message handlers for the edges in a
    known fashion, i.e. the message handler for an edge must be decorated with
    the edge name and the message model type.
    The message handler will be registered automatically and the message model
    will be used to validate the message content.

    Ex.:
        @dialogue._on_state_transition("edge_name", MessageModel)
        async def handler(ctx: Context, sender: str, message: MessageModel):
            pass

    A common practice is to add additional decorators to the pattern definition
    to simplify the usage of the dialogue class. This can be done by creating
    creating additional decorators that call the _on_state_transition method.
    Ex.:
        def on_init(model: type[Model]):
            return super()._on_state_transition("edge_name", model)

    and then use it like this:
        @pattern.on_init(MessageModel)
        async def handler(ctx: Context, sender: str, message: MessageModel):
            pass

    The current features include:
    - A graph representation of the dialogue, which is used to validate the
        sequence of messages.
    - Session handling which includes a session storage that contains all the
        messages that were exchanged between two participants.
    - Sessions will automatically be deleted after a certain amount of time.
    - Access to the dialogue history through ctx.dialogue (see Context class).
    """

    def __init__(
        self,
        name: str,
        storage: StorageAPI | None = None,
        nodes: list[Node] | None = None,
        edges: list[Edge] | None = None,
        timeout: int = DEFAULT_SESSION_TIMEOUT_IN_SECONDS,
        version: str | None = None,
        cleanup_interval: int = 1,
    ) -> None:
        self._name = name
        self._nodes = nodes or []
        self._edges = edges or []
        self._graph: dict[str, list[str]] = self._build_graph()  # by nodes
        self._rules: dict[str, list[str]] = self._build_rules()  # by edges
        self._digest_by_edge: dict[str, str] = {
            edge.name: Model.build_schema_digest(edge.model) if edge.model else ""
            for edge in self._edges
        }  # store the message models that are associated with an edge

        self._starter = self._build_starter()  # first message of the dialogue
        self._ender = self._build_ender()  # last message(s) of the dialogue

        self._timeout = timeout
        self._storage: StorageAPI = storage or KeyValueStore(
            f"{self._name}_dialogue_storage"
        )  # persistent session + message storage
        self._sessions: dict[UUID, list[Any]] = (
            self._load_storage()
        )  # volatile session + message storage
        self._states: dict[
            UUID, str
        ] = {}  # current state of the dialogue (as edge digest) per session

        if self._sessions:
            self._states = {
                session_id: session[-1]["schema_digest"]
                for session_id, session in self._sessions.items()
            }

        super().__init__(name=self._name, version=version)

        # if a model exists for an edge, register the handler automatically
        self._auto_add_message_handler()

        self.initialise_cleanup_task(cleanup_interval)

        # radical but effective
        self.on_message = lambda *args, **kwargs: None  # type: ignore
        self.on_query = lambda *args, **kwargs: None  # type: ignore

    @property
    def rules(self) -> dict[str, list[str]]:
        """
        Property to access the rules of the dialogue.

        Returns:
            dict[str, list[str]]: Dictionary of rules represented by edges.
        """
        return self._rules

    @property
    def nodes(self) -> list[Node]:
        return self._nodes

    @property
    def edges(self) -> list[Edge]:
        return self._edges

    def get_overview(self) -> dict:
        """
        Get an overview of the dialogue structure.

        Returns:
            dict: Manifest like representation of the dialogue structure.
        """
        return (
            {
                "nodes": [node.__dict__ for node in self._nodes],
                "edges": [
                    {
                        "name": edge.name,
                        "description": edge.description,
                        "parent": edge.parent.name if edge.parent else None,
                        "child": edge.child.name,
                        "model": edge.model.__name__ if edge.model else None,
                        "starter": edge.starter,
                        "ender": edge.ender,
                    }
                    for edge in self._edges
                ],
            }
            if self._nodes
            else {}
        )

    def _build_graph(self) -> dict[str, list[str]]:
        """
        Build the graph of the dialogue while showing the state relations.

        Returns:
            dict[str, list[str]]: List of states and their relations.
        """
        graph = {}
        for node in self._nodes:
            graph.setdefault(node.name, [])
        for edge in self._edges:
            if edge.parent:
                graph.setdefault(edge.parent.name, []).append(edge.child.name)
        return graph

    def _build_rules(self) -> dict[str, list[str]]:
        """
        Build the rules for the dialogue.
        Which replies are allowed after a certain message.

        Returns:
            dict[str, list[str]]: Rules for the dialogue.
        """
        out = {edge.name: [] for edge in self._edges}
        for edge in self._edges:
            for inner_edge in self._edges:
                if inner_edge.parent and inner_edge.parent.name == edge.child.name:
                    out[edge.name].append(inner_edge.name)
        return out

    def _build_starter(self) -> str:
        """Build the starting message of the dialogue."""
        starter_nodes = list(filter(lambda n: n.initial, self._nodes))
        # check if starter property has been set and if there is only one
        if len(starter_nodes) > 1:
            raise ValueError("Dialogue has more than one entry point!")

        edges_without_entry = list(filter(lambda e: e.parent is None, self._edges))
        if not edges_without_entry and starter_nodes:
            # if there is a starter node and no edge without parent we need to
            # validate if the graph is correct
            starters = list(filter(lambda e: e.parent is starter_nodes[0], self._edges))
            if starters:
                self._edges[self._edges.index(starters[0])].starter = True
                return starters[0].name
        if starter_nodes and edges_without_entry:
            warnings.warn(
                message="There is a starter node and an edge without parent present. "
                "The edge without a parent takes precedence!",
                category=SyntaxWarning,
                stacklevel=2,
            )
        if len(edges_without_entry) > 1:
            raise ValueError("Dialogue has more than one entry point!")
        if edges_without_entry:
            self._edges[self._edges.index(edges_without_entry[0])].starter = True
            return edges_without_entry[0].name
        raise ValueError("Dialogue has no entry point!")

    def _build_ender(self) -> set[str]:
        """Build the last message(s) of the dialogue and set final state."""
        for node, edges in self._graph.items():
            if not edges:
                self._nodes[
                    self._nodes.index(next(n for n in self._nodes if n.name == node))
                ].final = True
        enders = set()
        for edge in self._edges:
            if edge.child.final:
                enders.add(edge.name)
                edge.ender = True
        return enders

    def is_starter(self, digest: str) -> bool:
        """
        Return True if the digest is the starting message of the dialogue.
        False otherwise.
        """
        return self._digest_by_edge[self._starter] == digest

    def is_ender(self, digest: str) -> bool:
        """
        Return True if the digest is one of the last messages of the dialogue.
        False otherwise.
        """
        return digest in [self._digest_by_edge[edge] for edge in self._ender]

    def get_current_state(self, session_id: UUID) -> str:
        """Get the current state of the dialogue for a given session."""
        return self._states.get(session_id, "")

    def is_finished(self, session_id: UUID) -> bool:
        """
        Return True if the current state is (one of) the ending state(s).
        False otherwise.
        """
        return self.is_ender(self.get_current_state(session_id))

    def _pre_handle_hook(self, ctx: Context, sender: str, message: type[Model]) -> bool:
        schema_digest = Model.build_schema_digest(message)
        if not ctx.session:
            raise ValueError("Session ID must not be None!")
        is_valid = self.is_valid_message(ctx.session, schema_digest)
        if is_valid:
            if self.is_ender(schema_digest):
                self.update_state(schema_digest, ctx.session)
            self.add_message(
                session_id=ctx.session,
                message_type=self.models[schema_digest].__name__,
                schema_digest=schema_digest,
                sender=ctx.agent.address,
                receiver=sender,
                content=message.model_dump_json(),  # type: ignore
            )
        return is_valid

    def _post_handle_hook(self, ctx: Context, sender: str, msg_in: type[Model]) -> bool:
        if not ctx.session:
            raise ValueError("Session ID must not be None!")
        try:
            outbound_messages = ctx.outbound_messages  # type: ignore
        except AttributeError as exc:
            raise ValueError("Context must be an ExternalContext instance!") from exc

        if self.is_finished(ctx.session) or not outbound_messages:
            return True
        inbound_schema_digest = Model.build_schema_digest(msg_in)
        outbound_message_content, outbound_schema_digest = outbound_messages[sender][0]
        if not self.is_valid_reply(inbound_schema_digest, outbound_schema_digest):
            return False

        self.add_message(
            session_id=ctx.session,
            message_type=self.models[outbound_schema_digest].__name__,
            schema_digest=outbound_schema_digest,
            sender=ctx.agent.address,
            receiver=sender,
            content=outbound_message_content,
        )

        self.update_state(outbound_schema_digest, ctx.session)

        return True

    def _build_function_handler(self, edge: Edge) -> MessageCallback:
        if not edge.func:
            raise ValueError("No function handler set for edge!")

        @functools.wraps(edge.func)
        async def handler(ctx: Context, sender: str, message: type[Model]):
            # validate message first then execute handlers, finally update state
            if not self._pre_handle_hook(ctx, sender, message):
                return await ctx.send(
                    sender,
                    ErrorMessage(error=f"Unexpected message in dialogue: {message}"),
                )

            if edge.efunc:
                await edge.efunc(ctx, sender, message)
            result = await edge.func(ctx, sender, message)  # type: ignore

            if not self._post_handle_hook(ctx, sender, message):
                return MsgStatus(
                    status=DeliveryStatus.FAILED,
                    detail="Invalid dialogue reply",
                    destination=sender,
                    endpoint="",
                    session=ctx.session,
                )

            return result

        return handler  # type: ignore

    def _auto_add_message_handler(self) -> None:
        """Automatically add message handlers for edges with models."""
        for edge in self._edges:
            if edge.model and edge.func:
                self._add_message_handler(
                    model=edge.model,
                    func=self._build_function_handler(edge),
                    replies=None,  # no replies
                    allow_unverified=False,  # only verified
                )

    def update_state(self, digest: str, session_id: UUID) -> None:
        """
        Update the state of a dialogue session and create a new session
        if it does not exist.

        Args:
            digest (str): The digest of the message to update the state with.
            session_id (UUID): The ID of the session to update the state for.
        """
        self._states[session_id] = digest
        if session_id not in self._sessions:
            self._add_session(session_id)

    def _add_session(self, session_id: UUID) -> None:
        """Create a new session in the dialogue instance."""
        self._sessions[session_id] = []

    def cleanup_conversation(self, session_id: UUID) -> None:
        """Removes all messages related with the given session from the dialogue instance."""
        self._sessions.pop(session_id)
        self._remove_session_from_storage(session_id)

    def add_message(
        self,
        session_id: UUID,
        message_type: str,
        schema_digest: str,
        sender: str,
        receiver: str,
        content: JsonStr,
        **kwargs,
    ) -> None:
        """Add a message to the conversation of the given session within the dialogue instance."""
        if session_id is None:
            raise ValueError("Session ID must not be None!")
        if session_id not in self._sessions:
            self._add_session(session_id)
        self._sessions[session_id].append(
            {
                "message_type": message_type,
                "schema_digest": schema_digest,
                "sender": sender,
                "receiver": receiver,
                "message_content": content,
                "timestamp": datetime.timestamp(datetime.now()),
                "timeout": self._timeout,
                **kwargs,
            }
        )
        self._update_session_in_storage(session_id)

    def get_conversation(
        self, session_id: UUID, message_filter: str | None = None
    ) -> list[Any]:
        """
        Return the message history of the given session from the dialogue instance as
        list of DialogueMessage.
        This includes both sent and received messages.

        Args:
            session_id (UUID): The ID of the session to get the conversation for.
            message_filter (str | None): The name of the message type to filter for

        Returns:
            list(DialogueMessage): A list of all messages exchanged during the given session
            list(DialogueMessage): Only messages of type 'message_filter' (Model.__name__)
            from the given session
        """
        conversation = self._sessions.get(session_id, [])
        if message_filter is None:
            return conversation

        return [
            message
            for message in conversation
            if message["message_type"] == message_filter
        ]

    def get_edge(self, edge_name: str) -> Edge:
        """Return an edge from the dialogue instance."""
        return next(edge for edge in self._edges if edge.name == edge_name)

    def _resolve_mapping(self, msg_digest: str) -> str:
        """Resolve the mapping of a message digest to a message model."""
        return next(k for k, v in self._digest_by_edge.items() if v == msg_digest)

    def is_valid_message(self, session_id: UUID, msg_digest: str) -> bool:
        """
        Check if an incoming message is valid for a given session.

        Args:
            session_id (UUID): The ID of the session to check the message for.
            msg_digest (str): The digest of the message to check.

        Returns:
            bool: True if the message is valid, False otherwise.
        """
        if session_id not in self._sessions or len(self._sessions[session_id]) == 0:
            return self.is_starter(msg_digest)

        transition = self._resolve_mapping(msg_digest)
        current_state = self.get_current_state(session_id)
        current_state_mapped = self._resolve_mapping(current_state)
        allowed_msgs = self._rules.get(current_state_mapped, [])
        return transition in allowed_msgs

    def is_valid_reply(self, in_msg: str, out_msg: str) -> bool:
        """
        Check if a reply is valid for a given message.

        Args:
            in_msg (str): The digest of the message to check the reply for.
            out_msg (str): The digest of the reply to check.

        Returns:
            bool: True if the reply is valid, False otherwise.
        """
        return self._resolve_mapping(out_msg) in self._rules.get(
            self._resolve_mapping(in_msg), []
        )

    def is_included(self, msg_digest: str) -> bool:
        """
        Check if a message is included in the dialogue.

        Args:
            msg_digest (str): The digest of the message to check.

        Returns:
            bool: True if the message is included, False otherwise.
        """
        return msg_digest in self._digest_by_edge.values()

    def _load_storage(self) -> dict[UUID, list[Any]]:
        """Load the sessions from the storage."""
        cache: dict | None = self._storage.get(self._name)
        return (
            {UUID(session_id): session for session_id, session in cache.items()}
            if cache
            else {}
        )

    def _update_session_in_storage(self, session_id: UUID) -> None:
        """Update a session in the storage."""
        cache: dict = self._storage.get(self.name) or {}
        cache[str(session_id)] = self._sessions[session_id]
        self._storage.set(self.name, cache)

    def _remove_session_from_storage(self, session_id: UUID) -> None:
        """Remove a session from the storage."""
        cache: dict = self._storage.get(self.name) or {}
        session = str(session_id)
        if session in cache:
            cache.pop(session)
        self._storage.set(self.name, cache)

    def _update_transition_model(self, edge: Edge, model: type[Model]) -> None:
        """Update the message model for a transition."""
        self._digest_by_edge[edge.name] = Model.build_schema_digest(model)
        self._edges[self._edges.index(edge)].model = model

    def _on_state_transition(self, edge_name: str, model: type[Model]) -> Callable:
        """
        Main decorator to register a message handler for the dialogue.
        The Model in this case is the message model type but it refers to the
        Edge which is the transition between two states.

        Replies will not be validated based on the decorator parameters but
        on the graph and its rules instead.

        Args:
            edge (Edge): Edge object that represents the transition.
            model (type[Model]): Message model type.
        """
        if edge_name not in self._digest_by_edge:
            raise ValueError("Edge does not exist in the dialogue!")

        def decorator_on_state_transition(func: MessageCallback) -> MessageCallback:
            edge = self.get_edge(edge_name)
            edge.func = func
            handler = self._build_function_handler(edge)
            self._update_transition_model(edge, model)
            self._add_message_handler(model, handler, None, False)
            return handler

        return decorator_on_state_transition

    def manifest(self) -> dict[str, Any]:
        """
        This method will add the dialogue structure to the original manifest
        and recalculate the digest.
        """
        updated_manifest = super().manifest() | {
            "nodes": [node.__dict__ for node in self._nodes],
            "edges": [
                {
                    "name": edge.name,
                    "description": edge.description,
                    "parent": edge.parent.name if edge.parent else None,
                    "child": edge.child.name,
                    "model": edge.model.__name__ if edge.model else None,
                    "starter": edge.starter,
                    "ender": edge.ender,
                }
                for edge in self._edges
            ],
        }
        new_digest = Protocol.compute_digest(updated_manifest)
        updated_manifest["metadata"]["digest"] = new_digest
        return updated_manifest

    async def start_dialogue(
        self, ctx: Context, destination: str, message: Model
    ) -> list[MsgStatus]:
        """
        Start a dialogue with a message.

        Args:
            ctx (Context): The current message context
            destination (str): Either the agent address of the receiver or a protocol digest
            message (Model): The current message to send

        Raises:
            ValueError: If the dialogue is not started with the specified starting message.
        """
        message_schema_digest = Model.build_schema_digest(message)
        if not self.is_starter(message_schema_digest):
            raise ValueError(
                "A dialogue can only be started with the specified starting message"
            )

        status_list: list[MsgStatus] = []

        if destination.startswith("proto:"):
            status_list = await ctx.broadcast(destination, message)
        elif destination.startswith("agent"):
            status_list.append(await ctx.send(destination, message))
        else:
            raise ValueError("Invalid destination address")

        for status in status_list:
            if status.status == DeliveryStatus.FAILED:
                continue
            if status.session is None:
                raise ValueError("Session ID must not be None!")

            self.add_message(
                session_id=status.session,
                message_type=self.models[message_schema_digest].__name__,
                schema_digest=message_schema_digest,
                sender=ctx.agent.address,
                receiver=status.destination,
                content=message.model_dump_json(),
            )
            self.update_state(message_schema_digest, status.session)

        return status_list

    def initialise_cleanup_task(self, interval: int = 1) -> None:
        """
        Initialise the cleanup task.

        Deletes sessions that have not been used for a certain amount of time.
        The task runs every second so the configured timeout is currently
        measured in seconds as well (interval time * timeout parameter).
        Sessions with 0 as timeout will never be deleted.

        *Important*:
        - setting the interval above 1 will act as a multiplier
        - setting it to 0 will disable the cleanup task
        """
        if interval == 0:
            return

        @self.on_interval(interval)
        async def cleanup_dialogue(_ctx: Context):
            mark_for_deletion = []
            for session_id, session in self._sessions.items():
                timeout = session[-1]["timeout"]
                if (
                    timeout > 0
                    and datetime.fromtimestamp(session[-1]["timestamp"])
                    + timedelta(seconds=timeout)
                    < datetime.now()
                ):
                    mark_for_deletion.append(session_id)
            if mark_for_deletion:
                for session_id in mark_for_deletion:
                    self.cleanup_conversation(session_id)



import asyncio
import json
from datetime import datetime, timezone
from logging import Logger
from typing import Any

import uvicorn
from pydantic import BaseModel, ValidationError
from pydantic.v1 import ValidationError as ValidationErrorV1
from pydantic.v1.error_wrappers import ErrorWrapper
from requests.structures import CaseInsensitiveDict
from uagents_core.envelope import Envelope
from uagents_core.identity import is_user_address
from uagents_core.models import ERROR_MESSAGE_DIGEST, ErrorMessage, Model

from uagents.communication import enclose_response_raw
from uagents.config import DEFAULT_ENVELOPE_TIMEOUT_SECONDS, RESPONSE_TIME_HINT_SECONDS
from uagents.dispatch import dispatcher
from uagents.types import RestHandlerDetails, RestMethod
from uagents.utils import get_logger

HOST = "0.0.0.0"

RESERVED_ENDPOINTS = ["/submit", "/messages", "/agent_info", "/connect", "/disconnect"]


async def _read_asgi_body(receive):
    """Read the entire body of an ASGI message."""
    body = b""
    more_body = True

    while more_body:
        message = await receive()
        body += message.get("body", b"")
        more_body = message.get("more_body", False)

    return body


class ASGIServer:
    """ASGI server for receiving incoming envelopes."""

    def __init__(
        self,
        port: int,
        loop: asyncio.AbstractEventLoop,
        queries: dict[str, asyncio.Future],
        logger: Logger | None = None,
    ):
        """
        Initialize the ASGI server.

        Args:
            port (int): The port to listen on.
            loop (asyncio.AbstractEventLoop): The event loop to use.
            queries (dict[str, asyncio.Future]): The dictionary of queries to resolve.
            logger (Logger | None): The logger to use.
        """
        self._port = int(port)
        self._loop = loop
        self._queries = queries
        self._rest_handler_map: dict[
            tuple[str, RestMethod, str], RestHandlerDetails
        ] = {}
        self._logger = logger or get_logger("server")
        self._server: uvicorn.Server | None = None

    @property
    def server(self) -> uvicorn.Server | None:
        """Property to access the underlying uvicorn server."""
        return self._server

    def add_rest_endpoint(
        self,
        address: str,
        method: RestMethod,
        endpoint: str,
        request: type[Model] | None,
        response: type[Model | BaseModel],
    ):
        """Add a REST endpoint to the server."""
        self._rest_handler_map[(address, method, endpoint)] = RestHandlerDetails(
            method=method,
            endpoint=endpoint,
            request_model=request,
            response_model=response,
        )

    def has_rest_endpoint(self, method: RestMethod, endpoint: str) -> bool:
        """Check if the server has a REST endpoint registered."""
        if endpoint in RESERVED_ENDPOINTS:
            self._logger.warning(f"Endpoint {endpoint} is reserved")
            return True
        return any(
            meth == method and end == endpoint
            for sink, meth, end in self._rest_handler_map
        )

    def _get_rest_handler_details(
        self, method: RestMethod, endpoint: str
    ) -> dict[str, RestHandlerDetails]:
        handlers = {}
        for sink, meth, end in self._rest_handler_map:
            if meth == method and end == endpoint:
                handlers[sink] = self._rest_handler_map[(sink, meth, end)]
        return handlers

    async def _asgi_send(
        self,
        send,
        status_code: int = 200,
        headers: dict[str, str] | None = None,
        body: dict[str, Any] | ErrorWrapper | None = None,
    ):
        header = (
            [[k.encode(), v.encode()] for k, v in headers.items()]
            if headers is not None
            else [[b"content-type", b"application/json"]]
        )

        await send(
            {
                "type": "http.response.start",
                "status": status_code,
                "headers": header,
            }
        )

        if body is None:
            encoded_body = (
                b"{}" if [[b"content-type", b"application/json"]] in header else b""
            )
        else:
            encoded_body = json.dumps(body).encode()

        await send({"type": "http.response.body", "body": encoded_body})

    async def handle_readiness_probe(self, headers: CaseInsensitiveDict, send):
        """Handle a readiness probe sent via the HEAD method."""
        if b"x-uagents-address" not in headers:
            await self._asgi_send(
                send=send, headers={"x-uagents-status": "indeterminate"}
            )
        else:
            address = headers[b"x-uagents-address"].decode()  # type: ignore
            if not dispatcher.contains(address):
                await self._asgi_send(
                    send=send, headers={"x-uagents-status": "not-ready"}
                )
            else:
                await self._asgi_send(
                    send=send,
                    headers={
                        "x-uagents-status": "ready",
                        "x-uagents-response-time-hint": str(RESPONSE_TIME_HINT_SECONDS),
                    },
                )

    async def handle_missing_content_type(self, headers: CaseInsensitiveDict, send):
        """Handle missing content type header."""
        # if connecting from browser, return a 200 OK
        if b"user-agent" in headers:
            await self._asgi_send(send=send, body={"status": "OK - Agent is running"})
        else:  # otherwise, return a 400 Bad Request
            await self._asgi_send(
                send=send,
                status_code=400,
                body={"error": "missing header: content-type"},
            )

    async def serve(self):
        """Start the server."""
        config = uvicorn.Config(
            self,
            host=HOST,
            port=self._port,
            log_level="warning",
            forwarded_allow_ips="*",
            headers=[
                ("Access-Control-Allow-Origin", "*"),
                ("Access-Control-Allow-Methods", "GET, POST, OPTIONS"),
                ("Access-Control-Allow-Headers", "*"),
                ("Access-Control-Allow-Credentials", "false"),
            ],
        )
        self._server = uvicorn.Server(config)
        self._logger.info(
            f"Starting server on http://{HOST}:{self._port} (Press CTRL+C to quit)"
        )
        try:
            await self._server.serve()
        except (asyncio.CancelledError, KeyboardInterrupt):
            self._logger.info("Shutting down server...")

    async def _handle_rest(
        self,
        headers: CaseInsensitiveDict,
        handlers: dict[str, RestHandlerDetails],
        send,
        receive,
    ):
        raw_contents = await _read_asgi_body(receive)
        received_request: Model | None = None
        if len(handlers) > 1:
            if b"x-uagents-address" not in headers:
                await self._asgi_send(
                    send=send,
                    status_code=400,
                    body={
                        "error": "missing header: x-uagents-address",
                        "message": "Multiple handlers found for REST endpoint.",
                    },
                )
                return
            destination = headers[b"x-uagents-address"].decode()  # type: ignore
            rest_handler = handlers.get(destination)
        else:
            destination, rest_handler = handlers.popitem()

        if not rest_handler:
            await self._asgi_send(
                send=send, status_code=404, body={"error": "not found"}
            )
            return

        if rest_handler.method == "POST" and rest_handler.request_model is not None:
            if not raw_contents:
                await self._asgi_send(
                    send=send, status_code=400, body={"error": "No request body found"}
                )
                return

            try:
                received_request = rest_handler.request_model.model_validate_json(
                    raw_contents
                )
            except ValidationErrorV1 as err:
                e = dict(err.errors().pop())
                self._logger.debug(f"Failed to validate REST request: {e}")
                await self._asgi_send(send=send, status_code=400, body=e)
                return

        # get & call the handler
        handler_response: (
            dict[str, Any] | Model | None
        ) = await dispatcher.dispatch_rest(
            destination=destination,
            method=rest_handler.method,
            endpoint=rest_handler.endpoint,
            message=received_request,
        )

        # ensure the response is parsed as valid
        try:
            if not isinstance(handler_response, dict) and not isinstance(
                handler_response, rest_handler.response_model
            ):
                raise ValueError(
                    {"error": "Handler response must be a dict or a model"}
                )
            validated_response = rest_handler.response_model.model_validate(
                handler_response
            )
        except (ValidationErrorV1, ValueError) as err:
            self._logger.debug(f"Failed to validate REST response: {err}")
            await self._asgi_send(
                send=send,
                status_code=500,
                body={"error": "Handler response does not match response schema."},
            )
            return

        # return the validated response
        await self._asgi_send(send=send, body=validated_response.model_dump())

    async def __call__(self, scope, receive, send):  #  pylint: disable=too-many-branches
        """
        Handle an incoming ASGI message, dispatching the envelope to the appropriate handler,
        and waiting for any queries to be resolved.
        """
        scope_type = scope["type"]
        if scope_type == "lifespan" or scope_type != "http":
            return  # lifespan events not implemented and only handle http

        headers = CaseInsensitiveDict(scope.get("headers", {}))
        request_method = scope["method"]
        request_path = scope["path"]

        # Handle OPTIONS preflight request for CORS
        if request_method == "OPTIONS":
            await self._asgi_send(send=send, status_code=204, headers={})
            return

        # check if the request is for a REST endpoint
        handlers: dict[str, RestHandlerDetails] = self._get_rest_handler_details(
            method=request_method, endpoint=request_path
        )
        if handlers:
            if (
                request_path in set(RESERVED_ENDPOINTS) - {"/agent_info"}
                and "127.0.0.1" not in scope["client"]
            ):
                await self._asgi_send(
                    send=send, status_code=403, body={"error": "forbidden"}
                )
                return
            await self._handle_rest(headers, handlers, send, receive)
            return

        # check if the request is for agent communication and reject if not
        if request_path != "/submit":
            await self._asgi_send(
                send=send, status_code=404, body={"error": "not found"}
            )
            return

        if request_method == "HEAD":
            await self.handle_readiness_probe(headers, send)
            return

        if b"content-type" not in headers:
            await self.handle_missing_content_type(headers, send)
            return

        if b"application/json" not in headers[b"content-type"]:  # type: ignore
            await self._asgi_send(
                send=send, status_code=400, body={"error": "invalid content-type"}
            )
            return

        # read the entire payload
        raw_contents = await _read_asgi_body(receive)

        try:
            contents = json.loads(raw_contents.decode())
        except (AttributeError, UnicodeDecodeError, json.JSONDecodeError):
            await self._asgi_send(
                send=send, status_code=400, body={"error": "empty or invalid payload"}
            )
            return

        try:
            env = Envelope.model_validate(contents)
        except ValidationError:
            await self._asgi_send(
                send=send,
                status_code=400,
                body={"error": "contents do not match envelope schema"},
            )
            return

        expects_response = headers.get(b"x-uagents-connection") == b"sync"  # type: ignore

        if expects_response:
            # Add a future that will be resolved once the query is answered
            self._queries[env.sender] = asyncio.Future()

        if not is_user_address(env.sender):  # verify signature if sent from agent
            try:
                env.verify()
            except Exception as err:
                self._logger.warning(f"Failed to verify envelope: {err}")
                await self._asgi_send(
                    send=send, status_code=400, body={"error": str(err)}
                )
                return

        if not dispatcher.contains(address=env.target):
            await self._asgi_send(
                send=send, status_code=400, body={"error": "unable to route envelope"}
            )
            return

        await dispatcher.dispatch_msg(
            sender=env.sender,
            destination=env.target,
            schema_digest=env.schema_digest,
            message=env.decode_payload(),
            session=env.session,
        )

        # wait for any queries to be resolved
        if expects_response:
            timeout = (
                env.expires - datetime.now(timezone.utc).timestamp()
                if env.expires
                else DEFAULT_ENVELOPE_TIMEOUT_SECONDS
            )
            try:
                response_msg, schema_digest = await asyncio.wait_for(
                    self._queries[env.sender],
                    timeout,
                )
            except asyncio.TimeoutError:
                response_msg = ErrorMessage(
                    error="Query envelope expired"
                ).model_dump_json()
                schema_digest = ERROR_MESSAGE_DIGEST
            sender = env.target
            target = env.sender
            response = enclose_response_raw(
                json_message=response_msg,
                schema_digest=schema_digest,
                sender=sender,
                session=env.session,
                target=target,
            )
        else:
            response = "{}"

        await self._asgi_send(send=send, body=json.loads(response))



import logging
import time
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import (
    UUID4,
    BaseModel,
    ConfigDict,
    Field,
    field_serializer,
    field_validator,
)
from typing_extensions import Self
from uagents_core.envelope import Envelope
from uagents_core.models import Model

from uagents.config import (
    MESSAGE_HISTORY_MESSAGE_LIMIT,
    MESSAGE_HISTORY_RETENTION_SECONDS,
)
from uagents.storage import StorageAPI

if TYPE_CHECKING:
    from uagents.context import Context

JsonStr = str

IntervalCallback = Callable[["Context"], Awaitable[None]]
MessageCallback = Callable[["Context", str, Any], Awaitable[None]]
EventCallback = Callable[["Context"], Awaitable[None]]
WalletMessageCallback = Callable[["Context", Any], Awaitable[None]]

RestReturnType = dict[str, Any] | Model
RestGetHandler = Callable[["Context"], Awaitable[RestReturnType | None]]
RestPostHandler = Callable[["Context", Any], Awaitable[RestReturnType | None]]
RestHandler = RestGetHandler | RestPostHandler
RestMethod = Literal["GET", "POST"]
RestHandlerMap = dict[tuple[RestMethod, str], RestHandler]


AgentNetwork = Literal["mainnet", "testnet"]


class RestHandlerDetails(BaseModel):
    method: RestMethod
    endpoint: str
    request_model: type[Model] | None = None
    response_model: type[Model | BaseModel]


class AgentGeolocation(BaseModel):
    model_config = ConfigDict(strict=True, allow_inf_nan=False)
    latitude: Annotated[float, Field(ge=-90, le=90)]
    longitude: Annotated[float, Field(ge=-180, le=180)]
    radius: Annotated[float, Field(ge=0)] = 0

    @field_validator("latitude", "longitude")
    @classmethod
    def serialize_precision(cls, val: float) -> float:
        """
        Round the latitude and longitude to 6 decimal places.
        Equivalent to 0.11m precision.
        """
        return round(val, 6)


class AgentMetadata(BaseModel):
    """
    Model used to validate metadata for an agent.

    Framework specific fields will be added here to ensure valid serialization.
    Additional fields will simply be passed through.
    """

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
    )

    geolocation: AgentGeolocation | None = None


class MsgInfo(BaseModel):
    """
    Represents a message digest containing a message and its schema digest and sender.

    Attributes:
        message (Any): The message content.
        sender (str): The address of the sender of the message.
        schema_digest (str): The schema digest of the message.
    """

    message: Any
    sender: str
    schema_digest: str


class EnvelopeHistoryEntry(BaseModel):
    timestamp: int = Field(default_factory=lambda: int(time.time()))
    version: int
    sender: str
    target: str
    session: UUID4
    schema_digest: str
    protocol_digest: str | None = None
    payload: str | None = None

    @field_serializer("session")
    def serialize_session(self, session: UUID4, _info) -> JsonStr:
        return str(session)

    @classmethod
    def from_envelope(cls, envelope: Envelope) -> Self:
        return cls(
            version=envelope.version,
            sender=envelope.sender,
            target=envelope.target,
            session=envelope.session,
            schema_digest=envelope.schema_digest,
            protocol_digest=envelope.protocol_digest,
            payload=envelope.decode_payload(),
        )


class EnvelopeHistoryResponse(BaseModel):
    envelopes: list[EnvelopeHistoryEntry]


class SessionHistoryInfo(BaseModel):
    latest_timestamp: int
    message_count: int


class EnvelopeHistory:
    """
    Stores message history for an agent optionally using cache and/or storage.
    """

    def __init__(
        self,
        storage: StorageAPI,
        use_cache: bool = True,
        use_storage: bool = False,
        logger: logging.Logger | None = None,
        retention_period: int = MESSAGE_HISTORY_RETENTION_SECONDS,
        message_limit: int = MESSAGE_HISTORY_MESSAGE_LIMIT,
    ) -> None:
        """
        Initialize the message history.

        Args:
            storage (StorageAPI): The storage API to use.
            use_cache (bool): Whether to use the cache.
            use_storage (bool): Whether to use the storage.
            logger (logging.Logger | None): The logger to use.
            retention_period (int): The retention period in seconds.
            message_limit (int): The message limit.
        """
        self._cache: list[EnvelopeHistoryEntry] | None = [] if use_cache else None
        self._storage: StorageAPI | None = storage if use_storage else None
        self._logger = logger or logging.getLogger(__name__)
        self._retention_period = retention_period
        self._message_limit = message_limit

    def add_entry(self, entry: EnvelopeHistoryEntry) -> None:
        """
        Add an envelope entry to the message history.

        Args:
            entry (EnvelopeHistoryEntry): The entry to add.
        """
        if self._cache is not None:
            self._cache.append(entry)
        if self._storage is not None:
            key = self._get_key(entry.session)
            session_msgs: list[JsonStr] = self._storage.get(key) or []
            try:
                session_info = SessionHistoryInfo(
                    latest_timestamp=entry.timestamp,
                    message_count=len(session_msgs) + 1,
                )
                self._update_session_info(entry.session, session_info)
                session_msgs.append(entry.model_dump_json())
                self._storage.set(key, session_msgs)
            except RuntimeError as ex:
                self._logger.error(f"{ex.args[0]} Message will not be stored.")
        self.apply_retention_policy()

    def _update_session_info(
        self, session: UUID4, info_update: SessionHistoryInfo
    ) -> None:
        if self._storage is not None:
            all_sessions = self._storage.get("message-history:sessions") or {}
            total_msgs = 0
            for info_json in all_sessions.values():
                info = SessionHistoryInfo.model_validate_json(info_json)
                total_msgs += info.message_count
            if total_msgs >= self._message_limit:
                raise RuntimeError("Message history storage limit exceeded!")
            all_sessions.update({str(session): info_update.model_dump_json()})
            self._storage.set("message-history:sessions", all_sessions)

    def _get_key(self, session: UUID4 | str) -> str:
        return f"message-history:session:{str(session)}"

    def get_cached_messages(self) -> EnvelopeHistoryResponse:
        """
        Retrieve the cached message history.

        Returns:
            EnvelopeHistoryResponse: The cached message history.
        """
        if self._cache is None:
            raise ValueError("EnvelopeHistory cache is not set")
        return EnvelopeHistoryResponse(envelopes=self._cache)

    def get_session_messages(self, session: UUID4) -> list[EnvelopeHistoryEntry]:
        """
        Retrieve the message history for a given session.

        Args:
            session (UUID4): The session UUID.

        Returns:
            list[EnvelopeHistoryEntry]: The message history for the session.
        """
        if self._storage is None:
            raise ValueError("EnvelopeHistory storage is not set")
        session_msgs = self._storage.get(self._get_key(session)) or []
        return [EnvelopeHistoryEntry.model_validate_json(msg) for msg in session_msgs]

    def apply_retention_policy(self) -> None:
        """Remove entries older than retention period."""
        cutoff_time = time.time() - self._retention_period

        # apply retention policy to cache
        if self._cache is not None:
            for e in self._cache:
                if e.timestamp < cutoff_time:
                    self._cache.remove(e)
                else:
                    break
            if len(self._cache) > self._message_limit:
                self._cache = self._cache[-self._message_limit :]

        # apply retention policy to storage
        if self._storage is not None:
            all_sessions = self._storage.get("message-history:sessions") or {}
            sessions_to_remove = []
            for session, info_json in all_sessions.items():
                info = SessionHistoryInfo.model_validate_json(info_json)
                if info.latest_timestamp < cutoff_time:
                    self._storage.remove(self._get_key(session))
                    sessions_to_remove.append(session)
            for session in sessions_to_remove:
                del all_sessions[session]

            self._storage.set("message-history:sessions", all_sessions)
